/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "savage_driver.h"
#include "savage_vbe.h"
#include "savage_util.h"

#define iabs(a)	((int)(a)>0?(a):(-(a)))

Bool SavageSetupDevice(SavagePtr psav);
void SavageSetTextMode(SavagePtr psav);
Bool SavageSetVESAMode(SavagePtr psav, int n, int Refresh);
void SavageFreeBIOSModeTable(SavagePtr psav, SavageModeTablePtr* ppTable);
SavageModeTablePtr SavageGetBIOSModeTable(SavagePtr psav, int iDepth);
ModeStatus SavageMatchBiosMode(ScrnInfoPtr pScrn,int width,int height,int refresh,
                              unsigned int *vesaMode,unsigned int *newRefresh);
int SavageGetBIOSModes(SavagePtr psav,int iDepth,SavageModeEntryPtr s3vModeTable);
Bool SavageSetActiveDevice(SavagePtr psav,int activedev);

static Bool SavageQueryVBEInfo(SavagePtr psav);
static Bool SavageGetConnectionStatus(SavagePtr psav);
static Bool SavageGetDisplaysActive(SavagePtr psav);
static Bool SavageSetTvConf(SavagePtr psav,int TvConfig,int SelectMask);

Bool SavageSaveRestoreState(SavagePtr psav,Bool save);

static void SavageClearVM86Regs(xf86Int10InfoPtr pInt);
static Bool SavageX86Int10(xf86Int10InfoPtr pInt10);

Bool SavageSetupDevice(SavagePtr psav)
{
    if (!SavageQueryVBEInfo(psav))
        goto ExitError;
    if (!SavageGetConnectionStatus(psav))
        goto ExitError;
    if (!SavageGetDisplaysActive(psav))
        goto ExitError;
    return TRUE;
    
  ExitError:
    xf86Msg(X_ERROR,"SavageSetupDevice failed\n");
    return FALSE;
}
                       
Bool SavageQueryVBEInfo(SavagePtr psav)
{
	UnProtectCRTC();
    UnLockExtRegs();

    SavageClearVM86Regs(psav->pInt10);
    /* Query the VBE information */
    psav->pInt10->ax = S3_EXTBIOS_INFO;    
    psav->pInt10->bx = S3_VBE_INFO;
    
    if (!SavageX86Int10(psav->pInt10))
        goto Int10Error;

    psav->BiosVersion = psav->pInt10->bx;
    psav->videoRambytes = (psav->pInt10->si & 0xffff) << 16;
    /*
     *  Display device support (BIOS format, also used in Device Extension)
     *  These are shifted ( >> 16 ) values for BIOS Query Function 0
     */
    psav->DisplayOutSupport =(unsigned long)((psav->pInt10->si >> 16)
                                             & BIOS_DEVICE_MASK);
    
    psav->TvInfo.TvSizeX = 0;
    psav->TvInfo.TvSizeY = 0;       

    psav->TvInfo.TvStandardSupport = 0;
    if (psav->DisplayOutSupport & BIOS_TV_ONLY) {
        if (psav->DisplayOutSupport & BIOS_TV_NTSC) {
            psav->TvInfo.TvStandardSupport |= UT_TV_STD_NTSC;
        }
        if (psav->DisplayOutSupport & BIOS_TV_PAL) {
            psav->TvInfo.TvStandardSupport |= UT_TV_STD_PAL;
        }

        psav->DisplayOutSupport |= UT_DEVICE_TV;
        
        /*
         * cx:15~8:type of external TV encoder(Twister only)
         *       0: no TV encoder detected
         *       1: Chrontel CH7005
         *       2: Chrontel CH7009
         *       3: BrookTree BT868/869
         *     7~0: TV encoder address(Twister only)
         */
        if (psav->Chipset == S3_TWISTER) {
            switch (psav->pInt10->cx & 0xFF00) {
                case 0x0300:
                    /*Use common ID 0xB869 for all BrookTree-series TV chips.*/
                    psav->TvInfo.TvChip = TVChip_BT869;
                    break; 
                case 0x0200:
                    /* This ID is for CH7009, CH7010, CH7011, CH7012 */
                    psav->TvInfo.TvChip = TVChip_CH7009;
                    break; 
                case 0x0100:
                    /* This ID is for CH7005, CH7006. */
                    psav->TvInfo.TvChip = TVChip_CH7005; 
                    break; 
                case 0x0400:
                    /* This ID is for all VIA-series TV chips. */
                    psav->TvInfo.TvChip = TVChip_VT1621;  
                    break;
                case 0:   
                    psav->DisplayOutSupport &= (~UT_DEVICE_TV);  
                    break; 
            }
            
            if (psav->DisplayOutSupport & UT_DEVICE_TV) {
                psav->TvInfo.TvI2CAddress = psav->pInt10->cx & 0x00FF;
                switch (psav->TvInfo.TvChip) {
                    case TVChip_CH7009:
                        psav->TvInfo.TvSizeX = 800;
                        psav->TvInfo.TvSizeY = 600;
                        break;
                    case TVChip_CH7005:
                    case TVChip_BT869:
                    default:
                        psav->TvInfo.TvSizeX = 640;
                        psav->TvInfo.TvSizeY = 480;
                        break;
                }
            }
        }
    }
    
    xf86Msg(X_PROBED,"S3:support device %d.\n",psav->DisplayOutSupport);
    xf86Msg(X_PROBED,"S3:tv size %dx%d.\n",psav->TvInfo.TvSizeX,
            psav->TvInfo.TvSizeY);
    
    return TRUE;
    
  Int10Error:
    xf86Msg(X_ERROR,"SavageQueryVBEInfo failed\n");
    return FALSE;
}

Bool SavageGetConnectionStatus(SavagePtr psav)
{
    int connection = 0;
   
    if (psav->DisplayOutSupport == UT_DEVICE_CRT1) {
        /* CRT-only */
        xf86Msg(X_INFO,"S3: BIOS supports only CRT. Assume CRT connected.\n");
        psav->DisplayOutsConnected = UT_DEVICE_CRT1;
        return TRUE;
    }
    
	UnProtectCRTC();
    UnLockExtRegs();
    
    /*
     * We use Mobile BIOS for Twister, but CR6D definition is different from
     * desktop BIOS. Can't use CR6D for Twister.
     * Also use mobile bios for P4M266 & P4N266  the value in CR6D if not right!
     */
    if ((psav->Chipset != S3_TWISTER)
        && (psav->Chipset != S3_PROSAVAGEDDR)) {
        unsigned char    byCR6D;

        OUTREG8(CRT_ADDRESS_REG, 0x6d);
        byCR6D = INREG8(CRT_DATA_REG);
        
        if (byCR6D & 0x10)
            connection |= UT_DEVICE_CRT1;
        if (byCR6D & 0x20)
            connection |= UT_DEVICE_LCD;        
        if (byCR6D & 0x40)
            connection |= UT_DEVICE_TV;

        /*if nothing detected, set to CRT only */
        if (!(byCR6D & 0x70))
            connection |= UT_DEVICE_CRT1;

        psav->DisplayOutsConnected = connection;

        return TRUE;
    }

    SavageClearVM86Regs(psav->pInt10);
    /* Query the BIOS for the detected devices. */
    psav->pInt10->ax = S3_EXTBIOS_INFO;
    psav->pInt10->bx = S3_QUERY_ATTACHED;
    if (!SavageX86Int10(psav->pInt10))
        goto Int10Error;
    
    psav->DisplayOutsConnected  = (psav->pInt10->cx & BIOS_DEVICE_MASK);
    
    if (!psav->DisplayOutsConnected) {
        /*
        *  On desktop products, if no devices reported as connected,
        *  still assume at least a CRT.  This SHOULD have been caught
        *  based on support=CRT-only, but if BIOS reported support
        *  for CRT+[LCD or TV] and no devices were reported connected,
        *  we'll assume a CRT is there.
        */
        xf86Msg(X_ERROR,"S3: No devices reported connected!\n");
        psav->DisplayOutsConnected  = UT_DEVICE_CRT1;
    }
    xf86Msg(X_PROBED,"S3:connected device %d.\n",psav->DisplayOutsConnected);        
    return TRUE;

  Int10Error:
    xf86Msg(X_ERROR,"SavageGetConnectionStatus failed\n");
    return FALSE;
}


Bool
SavageGetDisplaysActive(SavagePtr psav)
{
	UnProtectCRTC();
    UnLockExtRegs();
    
    if (psav->DisplayOutSupport == UT_DEVICE_CRT1) {
        psav->DisplayOutsActive = UT_DEVICE_CRT1;
        return TRUE;
    }
    else {
        psav->DisplayOutsActive = UT_DEVICE_NONE;
    }

	UnProtectCRTC();
    UnLockExtRegs();
    
    SavageClearVM86Regs(psav->pInt10);
    psav->pInt10->ax = S3_EXTBIOS_INFO;    
    psav->pInt10->bx = S3_GET_ACTIVE_DISP;
    if (!SavageX86Int10(psav->pInt10))
        goto Int10Error;
    
    psav->DisplayOutsActive = (psav->pInt10->cx & BIOS_DEVICE_MASK);
    if ((psav->DisplayOutSupport & UT_DEVICE_TV)
        && (psav->DisplayOutsConnected & UT_DEVICE_TV)
        && (psav->TvOn)) {
        psav->DisplayOutsActive |= UT_DEVICE_TV;
    }
    
    /*
     *  If a TV is supported, set some defaults based on TV detection
     *  information, any saved information from the last boot, etc.
     *
     *  Set the appropriate method for enabling the user to view the full
     *  NT desktop: contraction or panning.
     *
     *  Most of the parameters we set for the TV will not take effect
     *  until the next BIOS call to Set Active Displays or Set Mode.
     */
    if ((psav->DisplayOutSupport & UT_DEVICE_TV)
        && (psav->DisplayOutsConnected & UT_DEVICE_TV)) {
        
        psav->pInt10->ax = S3_EXTBIOS_INFO;    
        psav->pInt10->bx = S3_GET_TV_CONFIG;
        if (!SavageX86Int10(psav->pInt10))
            goto Int10Error;
        
        switch (psav->pInt10->cx & 0x03) {
            case 0x00:
                psav->TvInfo.TvSignalType = UT_TV_SGNL_S_VIDEO;
                break;
            case 0x01:/* ???? */
            case 0x02:
            case 0x03:
                psav->TvInfo.TvSignalType = UT_TV_SGNL_COMPOSITE;
                break;
        }
        
        switch (psav->pInt10->cx & 0x0c) {
            case S3_TV_TYPE_NTSC_J:
            case S3_TV_TYPE_NTSC_US:
                psav->TvInfo.TvStandardInUse = UT_TV_STD_NTSC;                
                break;
            case S3_TV_TYPE_PAL:
                psav->TvInfo.TvStandardInUse = UT_TV_STD_PAL;
                break;
            default:
                xf86Msg(X_ERROR,"S3: S3BiosGetTvConfiguration, TVTECH_TYPE_UNKNOWN.\n");
                psav->TvInfo.TvStandardInUse = TVTECH_TYPE_UNKNOWN;
        }

        switch (psav->pInt10->cx & 0x20) {
            case 0x00:
                psav->TvInfo.TvUnderscanning = FALSE;
                psav->TvInfo.TvForcePanning = FALSE;
                break;
            case 0x10:
            case 0x30:
                /*
                 * if the panning bit is set, then regardless of the
                 * underscan bit, force panning is in effect
                 */
                psav->TvInfo.TvUnderscanning = FALSE;
                psav->TvInfo.TvForcePanning = TRUE;
                break;
            case 0x20:
                psav->TvInfo.TvUnderscanning = TRUE;
                psav->TvInfo.TvForcePanning = FALSE;
                break;
        }
    }
    
    xf86Msg(X_PROBED,"S3:actived device %d.\n",psav->DisplayOutsActive);
    return TRUE;
    
  Int10Error:
    xf86Msg(X_ERROR,"SavageGetDisplaysActive failed\n");
    return FALSE;
}

Bool SavageSaveRestoreState(SavagePtr psav,Bool save)
{
    SavageClearVM86Regs(psav->pInt10);
    
    if (!psav->StatInfo.statBuf) {
        int realOff;
        
        psav->pInt10->ax = BIOS_SVGA_STATE;
        psav->pInt10->dx = S3_GET_SVGA_BUF;
        psav->pInt10->cx = 0x0f;        
        if (!SavageX86Int10(psav->pInt10))
            goto ErrorExit;
        /* bx: count of 64k byte block needed */
        psav->StatInfo.pageCnt = (psav->pInt10->bx * 64 + 4095) / 4096;
        psav->StatInfo.statBuf = xf86Int10AllocPages(psav->pInt10,psav->StatInfo.pageCnt,&realOff);
        if (!psav->StatInfo.statBuf)
            goto ErrorExit;
        psav->StatInfo.realSeg = SEG_ADDR(realOff);
        psav->StatInfo.realOff = SEG_OFF(realOff);
    }
    
    /* save/resotre the original hardware state */
    psav->pInt10->ax = BIOS_SVGA_STATE;    
    if (save) 
        psav->pInt10->dx = S3_SAVE_SVGA_STATE;
    else
        psav->pInt10->dx = S3_RESTORE_SVGA_STATE;
    /*
     * cx: bit0:save/resotre VGA hardware state
     *     bit1:save/resotre bios data area
     *     bit2:save/resotre dac state
     *     bit3:save/resotre extend SVGA registers
     *     bit4:save/resotre MMIO registers
     */
    psav->pInt10->cx = 0x0f;
    psav->pInt10->es = psav->StatInfo.realSeg;
    psav->pInt10->bx = psav->StatInfo.realOff;
    if (!SavageX86Int10(psav->pInt10)) {
        xf86Msg(X_ERROR,"SavageSaveRestoreState:restore original state fail\n");
        goto ErrorExit;
    }
    
#if 1
    /* save/resotre the original mode */
    if (save) {
        psav->pInt10->ax = BIOS_GET_VBE_MODE;
        if (!SavageX86Int10(psav->pInt10))
            goto ErrorExit;
        psav->StatInfo.origMode = psav->pInt10->bx & 0xfff;
        xf86Msg(X_INFO,"original mode = 0x%x(bx=0x%x)\n",
                psav->StatInfo.origMode,psav->pInt10->bx);
    } else if (psav->StatInfo.origMode < 0x100) {
        /* for standard mode */
        psav->pInt10->ax = BIOS_SET_VBE_MODE;
        psav->pInt10->bx = psav->StatInfo.origMode;
        if (!SavageX86Int10(psav->pInt10)) {
            xf86Msg(X_ERROR,"SavageSaveRestoreState:restore original mode fail\n");
            goto ErrorExit;
        }
    }
#endif
    
    return TRUE;
    
ErrorExit:
    xf86Msg(X_ERROR,"SavageSaveRestoreState failed\n");
    return FALSE;
}


void
SavageSetTextMode(SavagePtr psav)
{
    SavageClearVM86Regs(psav->pInt10);
    psav->pInt10->ax = 0x83;
    SavageX86Int10(psav->pInt10);
}


Bool
SavageSetVESAMode(SavagePtr psav, int n, int refresh)
{
    unsigned char byte;

    xf86Msg(X_INFO,"SavageSetVESAMode:mode=0x%x,refresh=%dHZ\n",n,refresh);
    
    SavageClearVM86Regs(psav->pInt10);
    
    /* Establish the refresh rate for this mode. */
    psav->pInt10->ax = S3_EXTBIOS_INFO;
    psav->pInt10->bx = S3_SET_REFRESH;
    psav->pInt10->cx = n & 0x1ff;
    psav->pInt10->di = refresh & 0xffff;
    if (!SavageX86Int10(psav->pInt10)) {
        xf86Msg(X_ERROR,"SavageSetVESAMode:set refresh fail\n");
        goto ErrorExit;
    }

    /* SR01:turn off screen */
    OUTREG8 (SEQ_ADDRESS_REG,0x01);
    byte = INREG8(SEQ_DATA_REG) | 0x20;
    OUTREG8(SEQ_DATA_REG,byte);
    
    psav->pInt10->ax = BIOS_SET_VBE_MODE;
    psav->pInt10->bx = n;
    if (!SavageX86Int10(psav->pInt10)) {
        xf86Msg(X_ERROR,"SavageSetVESAMode:set vesa mode fail\n");
        goto ErrorExit;
    }
    
    return TRUE;
    
  ErrorExit:
    xf86Msg(X_ERROR,"SavageSetVESAMode failed\n");
    return FALSE;
}


SavageModeTablePtr
SavageGetBIOSModeTable(SavagePtr psav, int iDepth)
{
    int nModes = SavageGetBIOSModes(psav,iDepth,NULL);
    SavageModeTablePtr pTable;

    pTable = (SavageModeTablePtr) xcalloc(1, sizeof(SavageModeTableRec) +
                                          (nModes - 1) * sizeof(SavageModeEntry));
    if (pTable) {
        pTable->NumModes = nModes;
        SavageGetBIOSModes(psav, iDepth, pTable->Modes);
    }
    return pTable;
}


void
SavageFreeBIOSModeTable(SavagePtr psav, SavageModeTablePtr* ppTable)
{
    int i;
    SavageModeEntryPtr pMode = (*ppTable)->Modes;

    for (i = (*ppTable)->NumModes; i--;) {
        if (pMode->RefreshRate) {
            xfree( pMode->RefreshRate );
            pMode->RefreshRate = NULL;
        }
    }

    xfree(*ppTable);
}


ModeStatus SavageMatchBiosMode(ScrnInfoPtr pScrn,int width,int height,int refresh,
                              unsigned int *vesaMode,unsigned int *newRefresh)
{
    SavageModeEntryPtr pmt;
    Bool found = FALSE;
    SavagePtr psav = SAVPTR(pScrn);    
    int i,j;
    
    /* Scan through our BIOS list to locate the closest valid mode. */
    
    /*
     * If we ever break 4GHz clocks on video boards, we'll need to
     * change this.
     * refresh = (mode->Clock * 1000) / (mode->HTotal * mode->VTotal);
     * now we use VRefresh directyly,instead of by calculating from dot clock
     */
    for (i = 0, pmt = psav->ModeTable->Modes; 
         i < psav->ModeTable->NumModes;
         i++, pmt++) {
        if ((pmt->Width == width)
            /* this should be &&,but it seems that XFree86's int10 gets a zeor pmt->Height */ 
            || (pmt->Height == height)) {
            int jDelta = 99;
            int jBest = -1;

            if (vesaMode)
                *vesaMode = pmt->VesaMode;
            
            /* We have an acceptable mode.  Find a closest refresh rate. */            
            for (j = 0; j < pmt->RefreshCount; j++) {
                if (pmt->RefreshRate[j] == refresh) {
                    /* Exact match. */
                    jBest = j;
                    break;
                }
                else if (iabs(pmt->RefreshRate[j] - refresh) < jDelta) {
                    jDelta = iabs(pmt->RefreshRate[j] - refresh);
                    jBest = j;
                }
            }
            if (newRefresh)
                *newRefresh = pmt->RefreshRate[jBest];
            found = TRUE;
            break;
        }
    }

    if (found) {
        /* Success: we found a match in the BIOS. */
        xf86DrvMsg(pScrn->scrnIndex, X_PROBED, 
                   "SavageMatchBiosMode:chose mode %x at %dHz.\n", *vesaMode,*newRefresh);
        return MODE_OK;
        
    }
    else {
        xf86DrvMsg(pScrn->scrnIndex, X_PROBED, 
                   "No suitable BIOS mode found for %dx%d %dHz.\n",
                   width,height,refresh);
        return MODE_NOMODE;
    }
}



int
SavageGetBIOSModes(
    SavagePtr psav,
    int iDepth,
    SavageModeEntryPtr s3vModeTable )
{
    int iModeCount=0;
    int serialmodecall=0,modenumber;
    unsigned short type,depth;

    if (!psav->pVbe)
        return 0;
    
    SavageClearVM86Regs(psav->pInt10);
    
    do {
        psav->pInt10->ax = S3_EXTBIOS_INFO;/* extend bios functions */
        psav->pInt10->bx = S3_QUERY_MODELIST;
        psav->pInt10->dx = serialmodecall;
        if (!SavageX86Int10(psav->pInt10)) {
            goto Int10Error;
        }
        /* get new serial calling number */
        serialmodecall = psav->pInt10->dx & 0xffff;
        modenumber = (unsigned short)(psav->pInt10->cx & 0xffff);
        
        psav->pInt10->ax = S3_EXTBIOS_INFO;
        psav->pInt10->bx = 0x0302; /* get extended mode info */
        psav->pInt10->cx = modenumber;
        if (!SavageX86Int10(psav->pInt10)) {
            goto Int10Error;
        }
        depth = psav->pInt10->si & 0xff;
        type = (psav->pInt10->si >> 16) & 0xff;

        /* we only query these modes with same depth */
        if ((depth == iDepth)
            /* int10 module of XFree864.2 can't work properly,comment it and it always true */
            /*
            && (type == VBE_MODEL_256
                || type == VBE_MODEL_PACKED
                || type == VBE_MODEL_RGB)
            */   
            ) {
            /* This mode is a match. */
            iModeCount++;
            
            /* If we're supposed to fetch information, do it now. */
            if (s3vModeTable) {
                int iRefresh = 0,serialfreqcall=0;

                s3vModeTable->Width = psav->pInt10->di >> 16;
                s3vModeTable->Height = psav->pInt10->di & 0xffff;
                s3vModeTable->VesaMode = modenumber;
                
                /* Query the refresh rates at this mode. */
                do {
                    if ((iRefresh % 8) == 0) {
                        if (s3vModeTable->RefreshRate) {
                            s3vModeTable->RefreshRate = (unsigned char *)
                                xrealloc(s3vModeTable->RefreshRate,
                                         (iRefresh+8) * sizeof(unsigned char));
                        } else {
                            s3vModeTable->RefreshRate = (unsigned char *)
                                xcalloc(sizeof(unsigned char),(iRefresh+8));
                        }
                    }
                    

                    psav->pInt10->ax = S3_EXTBIOS_INFO;
                    psav->pInt10->bx = S3_QUERY_REFRESH;
                    psav->pInt10->cx = modenumber;
                    psav->pInt10->dx = serialfreqcall;
                    if (!SavageX86Int10(psav->pInt10))
                        goto Int10Error;
                    s3vModeTable->RefreshRate[iRefresh++] =  psav->pInt10->di;
                    serialfreqcall = psav->pInt10->dx;
                } while (serialfreqcall);

                s3vModeTable->RefreshCount = iRefresh;
                s3vModeTable++;
            }
        }
    } while (serialmodecall);/* 0 means the last time through the mode counting loop. */
    
    return iModeCount;
    
  Int10Error:
    xf86Msg(X_ERROR,"SavageGetBIOSModes failed\n");
    return -1;
}


Bool SavageSetActiveDevice(SavagePtr psav,int activedev)
{
    SavageClearVM86Regs(psav->pInt10);

    psav->pInt10->ax = S3_EXTBIOS_INFO;
    psav->pInt10->bx = S3_ALT_SET_ACTIVE_DISP;
    psav->pInt10->cx = activedev;
    if (!SavageX86Int10(psav->pInt10)) {
        xf86Msg(X_ERROR,"SavageSetActiveDevice failed\n");
        return FALSE;
    }
    return TRUE;
}


static Bool SavageSetTvConf(
    SavagePtr psav,
    int TvConfig,
    int SelectMask)
{
    UnLockExtRegs();
    
    SavageClearVM86Regs(psav->pInt10);
    psav->pInt10->bx = S3_SET_TV_CONFIG;
    psav->pInt10->ax = S3_EXTBIOS_INFO;
    psav->pInt10->cx = TvConfig;
    psav->pInt10->dx = SelectMask;
    if (SavageX86Int10 (psav->pInt10))
        return TRUE;
    else
        return FALSE;
}

static void SavageClearVM86Regs(xf86Int10InfoPtr pInt)
{
    pInt->ax = 0;
    pInt->bx = 0;
    pInt->cx = 0;
    pInt->dx = 0;
    pInt->si = 0;
    pInt->di = 0;
    pInt->es = 0;
    pInt->num = 0x10;
}

static Bool SavageX86Int10(xf86Int10InfoPtr pInt10)
{
    xf86ExecX86int10(pInt10);
    if ((pInt10->ax & 0xffff) == 0x004f)
        return TRUE;
    else {
        xf86Msg(X_ERROR,"xf86ExecX86int10 failed\n");        
        return FALSE;
    }
}
