/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#include <stdlib.h>
#include <stdio.h>

#include <GL/gl.h>

#include "mm.h"
#include "savagecontext.h"
#include "savagetex.h"
#include "savagelog.h"
#include "savageioctl.h"
#include "simple_list.h"
#include "enums.h"
#include "savage_bci.h"


GLuint gTileWidth[5] =
{
    64,    /* 4-bit */ 
    64,    /* 8-bit */
    64,    /* 16-bit */
    0,     /* 24-bit */
    32     /* 32-bit */
};

GLuint gTileHeight[5] = 
{
    64,   /* 4-bit */
    32,   /* 8-bit */
    16,   /* 16-bit */
    0,    /* 24-bit */
    16    /* 32-bit */
};

__inline GLuint GetTiledCoordinates8(GLuint iBufferWidth, GLint x, GLint y)
{
    GLint x10, x106, x52;
    GLint y20, y105, y43;
    GLuint uWidthInTiles;

    uWidthInTiles = (iBufferWidth + 63) >> 6;
    x10 = x & 0x3;
    x52 = (x & 0x3c) >> 2;
    x106 = (x & 0x7c0) >> 6;

    y20 = y & 0x7;
    y43 = (y & 0x18) >> 3;
    y105 = (y & 0x7e0) >> 5;
    
    return ( x10         |
            (y20 << 2)   |
	    (x52 << 5)   |
	    (y43 << 9)   |
	    ((y105 * uWidthInTiles) + x106) << 11 );
}

__inline GLuint GetTiledCoordinates16( GLint iBufferWidth,
                                              GLint x,
                                              GLint y )
{
    GLint  x106;
    GLint  x20;
    GLint  x53;
    GLint  y104;
    GLint  y20;
    GLint  y3;
    GLuint uiWidthInTiles;

    /*
    // calculating tiled address
    */

    uiWidthInTiles = (iBufferWidth + 63) >> 6;

    x20  =  x & 0x7;
    x53  = (x & 0x38 ) >> 3;
    x106 = (x & 0x7c0) >> 6;

    y20  =  y & 0x7;
    y3   = (y & 8    ) >> 3;
    y104 = (y & 0x7f0) >> 4;

    return( (x20 << 1)  |
            (y20 << 4)  |
            (x53 << 7)  |
            (y3  << 10) |
            ((y104 * uiWidthInTiles) + x106) << 11 );
}

__inline GLuint GetTiledCoordinates32( GLint iBufferWidth,
                                              GLint x,
                                              GLint y )
{
    GLint  x20;
    GLint  y20;
    GLuint uiWidthInTiles;
    GLint  x43;
    GLint  x105;
    GLint  y3;
    GLint  y104;

    /*
    // calculating tiled address
    */

    uiWidthInTiles = (iBufferWidth + 31) >> 5;

    x20  =  x & 0x7;
    x43  = (x & 0x18 ) >> 3;
    x105 = (x & 0x7e0) >> 5;

    y20  =  y & 0x7;
    y3   = (y & 8    ) >> 3;
    y104 = (y & 0x7f0) >> 4;

    return( (x20 << 2)  |
            (y20 << 5)  |
            (x43 << 8)  |
            (y3  << 10) |
            ((y104 * uiWidthInTiles) + x105) << 11 );
}

__inline GLuint GetTiledCoordinates( GLint iDepth,
                                            GLint iBufferWidth,
                                            GLint x,
                                            GLint y )
{
    /*
    // don't check for 4 since we only have 3 types of fb
    */

    if (iDepth == 16)
    {
        return( GetTiledCoordinates16( iBufferWidth, x, y ) );
    }
    else if (iDepth == 32)
    {
        return( GetTiledCoordinates32( iBufferWidth, x, y ) );
    }
    else
    {
        return( GetTiledCoordinates8( iBufferWidth, x, y ) );
    }
} 

__inline GLubyte * savageMakeTexel(savageTextureObjectPtr t, GLubyte * src, GLubyte * dst, GLuint level)
{
 
    GLushort u; 

    switch(t->image[level].baseFormat)
    {
        case GL_RGB:
	    if(t->image[level].internalFormat == TFT_ARGB1555)
	    {
        	*(GLushort *)dst = SAVAGEPACKCOLOR1555(src[0],src[1],src[2], 1);
                u = *(GLushort *)dst;
                src += 3;
	        return src;
            }
            else
            {
                if(t->image[level].internalFormat == TFT_ARGB4444)
		{
		    *(GLushort *)dst = SAVAGEPACKCOLOR4444(src[0],src[1],src[2], 0xf0);
                    src += 3;
	            return src;
                }
                else
		{
		    *(GLuint *)dst = SAVAGEPACKCOLOR8888(src[0],src[1],src[2], 0xff);
                    src += 3;
	            return src;
                }
            }
            break;
   
        case GL_RGBA:     
            switch (t->image[level].internalFormat)
            {
                case TFT_ARGB1555:
	            *(GLushort *)dst = SAVAGEPACKCOLOR1555(src[0],src[1],src[2], src[3]);
	            src += 4;
		    return src;
                case TFT_ARGB4444:
	            *(GLushort *)dst = SAVAGEPACKCOLOR4444(src[0],src[1],src[2], src[3]);
	            src += 4;
                    return src;	           
                case TFT_ARGB8888:
                    *(GLuint *)dst = SAVAGEPACKCOLOR8888(src[0],src[1],src[2], src[3]);
	            src += 4;
                    return src;
	    }
      	    break;

        case GL_LUMINANCE:
	    *dst = *src;
            src++;
            return src;
	    break;

        case GL_INTENSITY:
	    if(t->image[level].internalFormat == TFT_ARGB8888)
	    {
                *(GLuint *)dst = SAVAGEPACKCOLOR8888(src[0],src[0],src[0],src[0]);
                src++;
                return src;
	    }
            else
	    {
	        *(GLushort *)dst = SAVAGEPACKCOLOR4444(src[0],src[0],src[0],src[0]);
	        src++;
                return src;
	    } 
	    break;

        case GL_LUMINANCE_ALPHA:
            if(t->image[level].internalFormat == TFT_ARGB8888)
            {
	        *(GLuint *)dst = SAVAGEPACKCOLOR8888(src[0],src[0],src[0],src[1]);
	        src += 2;
                return src;
	    }
            else
            {
	        *(GLubyte *)dst = SAVAGEPACKCOLORA4L4(src[0],src[1]);
	        src += 2;
                return src;
	    }
            break;

        case GL_ALPHA:
            *dst = *src;
            src++;
            return src;
	    break;
        case GL_COLOR_INDEX:
            *dst = *src;
            src++;
            return src;
	    break;   
    }
    return src;
}


__inline void savageUploadImage(savageTextureObjectPtr t, GLuint level, GLuint startx, GLuint starty, GLuint reloc)
{
    GLuint uMaxTileWidth = gTileWidth[t->texelBytes]; 
    GLuint x, y, w, row, col;    
    const struct gl_texture_image *image = t->image[level].image;
    GLubyte * dst, * src, * pBuffer;
    GLint xAdd, yAdd;
    GLuint uRowSeparator, uChunk = MIN_TILE_CHUNK, uWrap;

  
    pBuffer = (GLubyte *)(t->BufAddr + t->image[level].offset);
    src = (GLubyte *)image->Data;
    x = startx;
    y = starty;
    w = image->Width;

    if(t->image[level].baseFormat ==  GL_COLOR_INDEX)
    {
        if(w < MIN_TILE_CHUNK)
	{
            w = MIN_TILE_CHUNK;
        }
        else
        {
	    if((w > 64 ) && (image->Height <= 16))
            {
	         reloc = GL_TRUE;
                 if(image->Height == 16)
		 {
		     uChunk = MIN_TILE_CHUNK << 1;
		 }
	    }
	}  

	if(!reloc & (w > (64 / 2)))
	{
	    for(row = 0; row < image->Height; row++)
	    {
	        for(col = 0; col < image->Width; col++)    
		{
		    dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, w, x + col, y + row);
                    src = savageMakeTexel(t, src, dst, level);
                }
	    }
	}
        else
        {
	    if(reloc & (w > 64))
	    {
	      uWrap = ((w + 63) >> 6) - 1;
	        for(row = 0; row < image->Height; row++)
	        {
	            for(col = 0; col < image->Width; col++)    
		    {
		        xAdd = (col / (4 * 64)) * 64 + col % 64;
                        yAdd = row + ((col / 64) & 3) * uChunk;
		        dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, 64, x + xAdd, y + yAdd);
                        src = savageMakeTexel(t, src, dst,level);
                    }
	        }
	    }
            else
	    {
	        uRowSeparator = 64 * MIN_TILE_CHUNK / w;
	        for(row = 0; row < image->Height; row++)
	        {
	            xAdd = (w * (row / MIN_TILE_CHUNK)) % 64;
                    yAdd = row % MIN_TILE_CHUNK + MIN_TILE_CHUNK * (row / uRowSeparator);
	            for(col = 0; col < image->Width; col++)    
		    {
		        dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, w, x + xAdd, y + yAdd);
                        src = savageMakeTexel(t, src, dst, level);
                    }
	        }
	    }
        }
    }
    else
    {
        if(w < MIN_TILE_CHUNK)
	{
            w = MIN_TILE_CHUNK;
        }
        else
        {
	    if((w > uMaxTileWidth ) && (image->Height <= 8))
            {
	         reloc = GL_TRUE;
	    }
	}
	if(!reloc & (w > uMaxTileWidth / 2))
	{
	    for(row = 0; row < image->Height; row++)
	    {
	        for(col = 0; col < image->Width; col++)    
		{
		    dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, w, x + col, y + row);
                    src = savageMakeTexel(t, src, dst, level);
                }
	    }
	}
        else
        {
	    if(reloc & (w > uMaxTileWidth))
	    {
	        for(row = 0; row < image->Height; row++)
	        {
	            for(col = 0; col < image->Width; col++)    
		    {
		        xAdd = (col / (2 * uMaxTileWidth)) * uMaxTileWidth + col % uMaxTileWidth;
                        yAdd = row + ((col / uMaxTileWidth) & 1) * MIN_TILE_CHUNK;
		        dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, w, x + xAdd, y + yAdd);
                        src = savageMakeTexel(t, src, dst, level);
                    }
	        }
	    }
            else
	    {
	        uRowSeparator = uMaxTileWidth * MIN_TILE_CHUNK / w;
	        for(row = 0; row < image->Height; row++)
	        {
                    yAdd = row % MIN_TILE_CHUNK + MIN_TILE_CHUNK * (row / uRowSeparator);
                    xAdd = (w * (row / MIN_TILE_CHUNK)) % uMaxTileWidth;
	            for(col = 0; col < image->Width; col++)    
		    {
		        dst = pBuffer + GetTiledCoordinates(t->texelBytes << 3, w, x + col + xAdd, y + yAdd);
                        src = savageMakeTexel(t, src, dst, level);
                    }
	        }
	    }
        }
    }
}        



__inline void  savageTileTex(savageTextureObjectPtr tex, GLuint level)
{
    GLuint uWidth, uHeight;
    GLint  xOffset, yOffset;
    GLint  xStart=0, yStart=0;
    GLint  minSize;
    GLuint xRepeat, yRepeat;
    GLuint startCol, startRow;
    GLuint reloc;

    const struct gl_texture_image *image = tex->image[level].image;
    
    reloc = GL_FALSE;
    uWidth = image->Width2;
    uHeight = image->Height2;

    if((uWidth > 4) || (uHeight > 4))
        minSize = MIN_TILE_CHUNK;
    else
        minSize = MIPMAP_CHUNK;
 
    if(image->Width >= minSize)
        xRepeat = 1;
    else
        xRepeat = minSize / image->Width;

    if(image->Height >= minSize)
        yRepeat = 1;
    else
    {
        yRepeat = minSize / image->Height;
        if(minSize == MIN_TILE_CHUNK)
            reloc = GL_TRUE;
    }
  
    if(((uWidth < 4) && (uHeight < 4)) && (tex->texelBytes >= 2))
    { 
        if((uWidth == 2) || (uHeight == 2))
        {
            xStart = 4;
            yStart = 0;
        }

        else
        {
            xStart = 4;
            yStart = 4;
        }
    }
    for(xOffset = 0; xOffset < xRepeat; xOffset++)
    {
        for(yOffset = 0; yOffset < yRepeat; yOffset++)
        {
	    startCol = image->Width * xOffset + xStart;
            startRow = image->Height * yOffset + yStart;
            savageUploadImage(tex,level, startCol, startRow, reloc);
	}
    }
}

static void savageSetTexWrapping(savageTextureObjectPtr tex, GLenum s, GLenum t)
{
    tex->texParams.sWrapMode = s;
    tex->texParams.tWrapMode = t;
}

static void savageSetTexFilter(savageContextPtr imesa, 
			     savageTextureObjectPtr t, 
			     GLenum minf, GLenum magf)
{
   t->texParams.minFilter = minf;
   t->texParams.magFilter = magf;
}


/* Need a fallback ?
 */
static void savageSetTexBorderColor(savageTextureObjectPtr t, GLubyte color[4])
{
/*    t->Setup[SAVAGE_TEXREG_TEXBORDERCOL] =  */
      t->texParams.boarderColor = SAVAGEPACKCOLOR8888(color[0],color[1],color[2],color[3]); 
}



static void ReplicateMesaTexState(savageContextPtr imesa,
				  savageTextureObjectPtr t,
                                  struct gl_texture_object *mesatex)
{
   savageSetTexWrapping(t,mesatex->WrapS,mesatex->WrapT);
   savageSetTexFilter(imesa, t,mesatex->MinFilter,mesatex->MagFilter);
   savageSetTexBorderColor(t,mesatex->BorderColor);
}

static savageTextureObjectPtr savageCreateTexObj(savageContextPtr imesa,
					     struct gl_texture_object *tObj)
{
   savageTextureObjectPtr t;
   GLuint offset, width, pitch, i, textureFormat, log_pitch, baseFormat;
   struct gl_texture_image *image;

   image = tObj->Image[ 0 ];
   if ( !image ) {
      fprintf(stderr, "no image at level zero - not creating texobj\n");
      return 0;
   }
   
   t = (savageTextureObjectPtr) calloc(1,sizeof(*t));
   if (!t)
      return 0;

   switch( image->Format ) {
   case GL_R3_G3_B2:
   case GL_RGB4:
      t->texelBytes = 2;
      baseFormat = GL_RGB;
      textureFormat = TFT_ARGB4444;
      break;

   case GL_RGB5:
      t->texelBytes = 2;
      baseFormat = GL_RGB;
      textureFormat = TFT_ARGB1555;
      break;
   case GL_RGB:
   case GL_RGB8:
   case GL_RGB10:
   case GL_RGB12:
   case GL_RGB16:
      t->texelBytes = 4;
      baseFormat = GL_RGB;
      textureFormat = TFT_ARGB8888;
      break;

   case GL_LUMINANCE:
   case GL_LUMINANCE4:
   case GL_LUMINANCE8:
   case GL_LUMINANCE12:
   case GL_LUMINANCE16:
      t->texelBytes = 1;
      baseFormat =  GL_LUMINANCE;
      textureFormat = TFT_L8;
      break;
   case GL_ALPHA:
   case GL_ALPHA4:
   case GL_ALPHA8:
   case GL_ALPHA12:
   case GL_ALPHA16:
      baseFormat = GL_ALPHA;
      t->texelBytes = 1;
      textureFormat = TFT_A8;
      break;

   case GL_LUMINANCE_ALPHA:
   case GL_LUMINANCE4_ALPHA4:
   case GL_LUMINANCE6_ALPHA2:
   case GL_LUMINANCE8_ALPHA8:
   case GL_LUMINANCE12_ALPHA4:
   case GL_LUMINANCE12_ALPHA12:
   case GL_LUMINANCE16_ALPHA16:
      t->texelBytes = 1;
      baseFormat =  GL_LUMINANCE_ALPHA;
      if(image->Format == GL_LUMINANCE4_ALPHA4)
          textureFormat = TFT_A4L4;
      else
      {
	  t->texelBytes = 4;
          textureFormat = TFT_ARGB8888;
      }
      break;

   case GL_INTENSITY:
   case GL_INTENSITY4:
   case GL_INTENSITY8:
   case GL_INTENSITY12:
   case GL_INTENSITY16:
      t->texelBytes = 2;
      baseFormat = GL_INTENSITY;
      if(image->Format == GL_INTENSITY4)
	  textureFormat = TFT_ARGB4444;
      else
      {
          t->texelBytes = 4;
          textureFormat = TFT_ARGB8888;
      }
      break;

 /*  case GL_RGBA:*/
   case GL_RGBA2:
   case GL_RGBA4:
      t->texelBytes = 2;
      baseFormat = GL_RGBA;
      textureFormat = TFT_ARGB4444;
      break;
   case GL_RGB5_A1:
      t->texelBytes = 2;
      baseFormat = GL_RGBA;
      textureFormat = TFT_ARGB1555;
      break;
   case GL_RGBA:
   case GL_RGBA8:
   case GL_RGBA12:
   case GL_RGBA16:
   case GL_RGB10_A2:
      t->texelBytes = 4;
      baseFormat = GL_RGBA;
      textureFormat = TFT_ARGB8888;
      break;

   case GL_COLOR_INDEX:
   case GL_COLOR_INDEX1_EXT:
   case GL_COLOR_INDEX2_EXT:
   case GL_COLOR_INDEX4_EXT:
   case GL_COLOR_INDEX8_EXT:
   case GL_COLOR_INDEX12_EXT:
   case GL_COLOR_INDEX16_EXT:
      baseFormat = GL_COLOR_INDEX;
      textureFormat = GL_COLOR_INDEX;
      t->texelBytes = 1;
      break;
   default:
      free( t );      
      return 0;	
   }


   /* Figure out the size now (and count the levels).  Upload won't be done
    * until later.
    */ 
   width = image->Width * t->texelBytes;
   for (pitch = 2, log_pitch=1 ; pitch < width ; pitch *= 2 )
      log_pitch++;
   
   t->dirty_images = 0;
   t->bound = 0;
   
   t->Pitch = pitch;
   offset = 0;

   for ( i = 0 ; i < SAVAGE_TEX_MAXLEVELS && tObj->Image[i] ; i++ ) {
      t->image[i].image = tObj->Image[i];
      t->image[i].offset = offset;
      t->image[i].internalFormat = textureFormat;
      t->image[i].baseFormat = baseFormat;
      t->dirty_images |= (1<<i);
      offset += t->image[i].image->Height * pitch;
      pitch = pitch >> 1;
   }

   t->totalSize = offset;
   t->max_level = i-1;
   t->min_level = 0;
   t->globj = tObj;
   t->age = 0;


   t->current_unit = 0;

   ReplicateMesaTexState(imesa, t,tObj);
   tObj->DriverData = t;
   imesa->dirty |= SAVAGE_UPLOAD_CTX;
   make_empty_list( t );
   return t;
}

void savageDestroyTexObj(savageContextPtr imesa, savageTextureObjectPtr t)
{
   if (!t) return;

   /* This is sad - need to sync *in case* we upload a texture
    * to this newly free memory...
    */
   if (t->MemBlock) {
      mmFreeMem(t->MemBlock);
      t->MemBlock = 0;

      if (t->age > imesa->dirtyAge)
	 imesa->dirtyAge = t->age;
   }

   if (t->globj)
      t->globj->DriverData = 0;

   if (t->bound)
      imesa->CurrentTexObj[t->bound - 1] = 0; 

   remove_from_list(t);
   free(t);
}


static void savageSwapOutTexObj(savageContextPtr imesa, savageTextureObjectPtr t)
{
   if (t->MemBlock) {
      mmFreeMem(t->MemBlock);
      t->MemBlock = 0;      

      if (t->age > imesa->dirtyAge)
	 imesa->dirtyAge = t->age;
   }

   t->dirty_images = ~0;
   move_to_tail(&(imesa->SwappedOut), t);
}



/* Upload an image from mesa's internal copy.
 */
static void savageUploadTexLevel( savageTextureObjectPtr t, int level )
{
   const struct gl_texture_image *image = t->image[level].image;


   /* Need triangle (rather than pixel) fallbacks to simulate this using
    * normal textured triangles.
    *
    * DO THIS IN DRIVER STATE MANAGMENT, not hardware state.
    * 
    */

   if(image->Border != 0) 
       savageError("Not supported texture border %d.\n", (int) image->Border);
   
   
   switch(t->image[level].baseFormat)
   {
       case GL_RGB:
       case GL_RGBA:
       case GL_LUMINANCE:
       case GL_INTENSITY:
       case GL_LUMINANCE_ALPHA:
       case GL_ALPHA:
       case GL_COLOR_INDEX:
           savageTileTex(t, level);
           break;
       default:
           savageError("Not supported texture format %d.\n", (GLint) t->image[level].baseFormat);
   }
}



void savagePrintLocalLRU( savageContextPtr imesa , GLuint heap) 
{
   savageTextureObjectPtr t;
   int sz = 1 << (imesa->savageScreen->logTextureGranularity[heap]);
   
   foreach( t, &imesa->TexObjList[heap] ) {
      if (!t->globj)
	 fprintf(stderr, "Placeholder %d at %x sz %x\n", 
		 t->MemBlock->ofs / sz,
		 t->MemBlock->ofs,
		 t->MemBlock->size);      
      else
	 fprintf(stderr, "Texture (bound %d) at %x sz %x\n", 
		 t->bound,
		 t->MemBlock->ofs,
		 t->MemBlock->size);      

   }
}

void savagePrintGlobalLRU( savageContextPtr imesa , GLuint heap)
{
   int i, j;

   drm_savage_tex_region_t *list = imesa->sarea->texList[heap];
   

   for (i = 0, j = SAVAGE_NR_TEX_REGIONS ; i < SAVAGE_NR_TEX_REGIONS ; i++) {
      fprintf(stderr, "list[%d] age %d next %d prev %d\n",
	      j, list[j].age, list[j].next, list[j].prev);
      j = list[j].next;
      if (j == SAVAGE_NR_TEX_REGIONS) break;
   }
   
   if (j != SAVAGE_NR_TEX_REGIONS)
      fprintf(stderr, "Loop detected in global LRU\n");
       for (i = 0 ; i < SAVAGE_NR_TEX_REGIONS ; i++) 
       {
          fprintf(stderr,"list[%d] age %d next %d prev %d\n",
          i, list[i].age, list[i].next, list[i].prev);
       }
}


void savageResetGlobalLRU( savageContextPtr imesa, GLuint heap )
{
    drm_savage_tex_region_t *list = imesa->sarea->texList[heap];
   int sz = 1 << imesa->savageScreen->logTextureGranularity[heap];
   int i;

   /* (Re)initialize the global circular LRU list.  The last element
    * in the array (SAVAGE_NR_TEX_REGIONS) is the sentinal.  Keeping it
    * at the end of the array allows it to be addressed rationally
    * when looking up objects at a particular location in texture
    * memory.  
    */
   for (i = 0 ; (i+1) * sz <= imesa->savageScreen->textureSize[heap]; i++) {
      list[i].prev = i-1;
      list[i].next = i+1;
      list[i].age = 0;
   }

   i--;
   list[0].prev = SAVAGE_NR_TEX_REGIONS;
   list[i].prev = i-1;
   list[i].next = SAVAGE_NR_TEX_REGIONS;
   list[SAVAGE_NR_TEX_REGIONS].prev = i;
   list[SAVAGE_NR_TEX_REGIONS].next = 0;
   imesa->sarea->texAge[heap] = 0;
}


static void savageUpdateTexLRU( savageContextPtr imesa, savageTextureObjectPtr t ) 
{
   int i;
   int heap = t->heap;
   int logsz = imesa->savageScreen->logTextureGranularity[heap];
   int start = t->MemBlock->ofs >> logsz;
   int end = (t->MemBlock->ofs + t->MemBlock->size - 1) >> logsz;
   drm_savage_tex_region_t *list = imesa->sarea->texList[heap];
   
   imesa->texAge[heap] = ++imesa->sarea->texAge[heap];

   /* Update our local LRU
    */
   move_to_head( &(imesa->TexObjList[heap]), t );

   /* Update the global LRU
    */
   for (i = start ; i <= end ; i++) {

      list[i].in_use = 1;
      list[i].age = imesa->texAge[heap];

      /* remove_from_list(i)
       */
      list[(unsigned)list[i].next].prev = list[i].prev;
      list[(unsigned)list[i].prev].next = list[i].next;
      
      /* insert_at_head(list, i)
       */
      list[i].prev = SAVAGE_NR_TEX_REGIONS;
      list[i].next = list[SAVAGE_NR_TEX_REGIONS].next;
      list[(unsigned)list[SAVAGE_NR_TEX_REGIONS].next].prev = i;
      list[SAVAGE_NR_TEX_REGIONS].next = i;
   }
}


/* Called for every shared texture region which has increased in age
 * since we last held the lock.
 *
 * Figures out which of our textures have been ejected by other clients,
 * and pushes a placeholder texture onto the LRU list to represent 
 * the other client's textures.  
 */
void savageTexturesGone( savageContextPtr imesa,
		       GLuint heap,
		       GLuint offset, 
		       GLuint size,
		       GLuint in_use ) 
{
   savageTextureObjectPtr t, tmp;
   
   foreach_s ( t, tmp, &imesa->TexObjList[heap] ) {

      if (t->MemBlock->ofs >= offset + size ||
	  t->MemBlock->ofs + t->MemBlock->size <= offset)
	 continue;

      /* It overlaps - kick it off.  Need to hold onto the currently bound
       * objects, however.
       */
      if (t->bound)
	 savageSwapOutTexObj( imesa, t );
      else
	 savageDestroyTexObj( imesa, t );
   }

   
   if (in_use) {
      t = (savageTextureObjectPtr) calloc(1,sizeof(*t));
      if (!t) return;

      t->heap = heap;
      t->MemBlock = mmAllocMem( imesa->texHeap[heap], size, 0, offset);      
      if(!t->MemBlock)
      {
          free(t);
          return;
      }
      insert_at_head( &imesa->TexObjList[heap], t );
   }
}





/* This is called with the lock held.  May have to eject our own and/or
 * other client's texture objects to make room for the upload.
 */
int savageUploadTexImages( savageContextPtr imesa, savageTextureObjectPtr t )
{
   int heap;
   int i;
   int ofs;
   
   heap = t->heap = SAVAGE_CARD_HEAP;

   /* Do we need to eject LRU texture objects?
    */
   if (!t->MemBlock) {
      while (1)
      {
	 t->MemBlock = mmAllocMem( imesa->texHeap[heap], t->totalSize, 12, 0 ); 
	 if (t->MemBlock)
	    break;
	 else
	 {
	     heap = t->heap = SAVAGE_AGP_HEAP;
	     t->MemBlock = mmAllocMem( imesa->texHeap[heap], t->totalSize, 12, 0 ); 
	     
	     if (t->MemBlock)
	         break;
	 }

	 if (imesa->TexObjList[heap].prev->bound) {
  	    fprintf(stderr, "Hit bound texture in upload\n"); 
	    savagePrintLocalLRU( imesa,heap );
	    return -1;
	 }

	 if (imesa->TexObjList[heap].prev == &(imesa->TexObjList[heap])) {
 	    fprintf(stderr, "Failed to upload texture, sz %d\n", t->totalSize);
	    mmDumpMemInfo( imesa->texHeap[heap] );
	    return -1;
	 }
	 
	 savageDestroyTexObj( imesa, imesa->TexObjList[heap].prev );
      }
 
      ofs = t->MemBlock->ofs;
      t->texParams.hwPhysAddress = imesa->savageScreen->textureOffset[heap] + ofs;
      t->BufAddr = (char *)((GLuint) imesa->savageScreen->texVirtual[heap] + ofs);
      imesa->dirty |= SAVAGE_UPLOAD_CTX;
   }

   /* Let the world know we've used this memory recently.
    */
   savageUpdateTexLRU( imesa, t );

   if (t->dirty_images) {
      if (SAVAGE_DEBUG & DEBUG_VERBOSE_LRU)
	 fprintf(stderr, "*");

      for (i = t->min_level ; i <= t->max_level ; i++)
	 if (t->dirty_images & (1<<i)) 
	    savageUploadTexLevel( t, i );
   }


   t->dirty_images = 0;
   return 0;
}

static void savageTexSetUnit( savageTextureObjectPtr t, GLuint unit )
{
   if (t->current_unit == unit) return;

   t->current_unit = unit;
}




static void savageUpdateTex0State( GLcontext *ctx )
{
   savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
   struct gl_texture_object	*tObj;
   savageTextureObjectPtr t;
   GLuint format;
   Reg_TexCtrl TexCtrl;
   Reg_TexBlendCtrl TexBlendCtrl;
   Reg_TexDescr TexDescr; 

   /* disable */
   
   if (ctx->Texture.Unit[0].ReallyEnabled == 0) {
      imesa->Registers.TexDescr.ni.tex0En = GL_FALSE;
      imesa->Registers.TexBlendCtrl[0].ui = TBC_NoTexMap;
      imesa->Registers.TexCtrl[0].ui = 0x20f040;
      imesa->Registers.TexAddr[0].ui = 0;
      imesa->Registers.changed.ni.fTex0BlendCtrlChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTex0AddrChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTexDescrChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTex0CtrlChanged = GL_TRUE;
      return;
   }

   tObj = ctx->Texture.Unit[0].Current;
   if (ctx->Texture.Unit[0].ReallyEnabled != TEXTURE0_2D ||
       tObj->Image[tObj->BaseLevel]->Border > 0) {
      /* 1D or 3D texturing enabled, or texture border - fallback */
      imesa->Fallback |= SAVAGE_FALLBACK_TEXTURE;
      return;
   }

   /* Do 2D texture setup */


   t = tObj->DriverData;
   if (!t) {
      t = savageCreateTexObj( imesa, tObj );
      if (!t)
         return;
   }

   if (t->current_unit != 0)
      savageTexSetUnit( t, 0 );
    
   imesa->CurrentTexObj[0] = t;
   t->bound = 1;

   if (t->dirty_images) 
       savageUploadTexImages(imesa, imesa->CurrentTexObj[0]); 
   
   if (t->MemBlock)
      savageUpdateTexLRU( imesa, t );
  
   TexDescr.ui     = imesa->Registers.TexDescr.ui & ~0x01000000;
   TexCtrl.ui      = imesa->Registers.TexCtrl[0].ui;
   TexBlendCtrl.ui = imesa->Registers.TexBlendCtrl[0].ui;


   format = t->image[0].baseFormat;

   switch (ctx->Texture.Unit[0].EnvMode) {
   case GL_REPLACE:
      TexCtrl.ni.clrArg1Invert = GL_FALSE;
      switch(format)
      {
          case GL_LUMINANCE:
          case GL_RGB:
               TexBlendCtrl.ui = TBC_Decal;
               break;

          case GL_LUMINANCE_ALPHA:
          case GL_RGBA:
          case GL_INTENSITY:
               TexBlendCtrl.ui = TBC_Copy;
               break;

          case GL_ALPHA:
               TexBlendCtrl.ui = TBC_CopyAlpha;
               break;
      }
       __HWEnvCombineSingleUnitScale(imesa, 0, 0, &TexBlendCtrl);
      break;

    case GL_DECAL:
        TexCtrl.ni.clrArg1Invert = GL_FALSE;
        switch (format)
        {
            case GL_RGB:
            case GL_LUMINANCE:
                TexBlendCtrl.ui = TBC_Decal;
                break;

            case GL_RGBA:
            case GL_INTENSITY:
            case GL_LUMINANCE_ALPHA:
                TexBlendCtrl.ui = TBC_DecalAlpha;
                break;

            /*
             GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_ALPHA, GL_INTENSITY
             are undefined with GL_DECAL
            */

            case GL_ALPHA:
                TexBlendCtrl.ui = TBC_CopyAlpha;
                break;
        }
        __HWEnvCombineSingleUnitScale(imesa, 0, 0, &TexBlendCtrl);
        break;

    case GL_MODULATE:
        TexCtrl.ni.clrArg1Invert = GL_FALSE;
        TexBlendCtrl.ui = TBC_ModulAlpha;
        __HWEnvCombineSingleUnitScale(imesa, 0, 0, &TexBlendCtrl);
        break;

    case GL_BLEND:

        switch (format)
        {
            case GL_ALPHA:
                TexBlendCtrl.ui = TBC_ModulAlpha;
                TexCtrl.ni.clrArg1Invert = GL_FALSE;
                break;

            case GL_LUMINANCE:
            case GL_RGB:
                TexBlendCtrl.ui = TBC_Blend0;
                TexDescr.ni.tex1En = GL_TRUE;
                TexDescr.ni.texBLoopEn = GL_TRUE;
                TexDescr.ni.tex1Width  = TexDescr.ni.tex0Width;
                TexDescr.ni.tex1Height = TexDescr.ni.tex0Height;
                TexDescr.ni.tex1Fmt = TexDescr.ni.tex0Fmt;

                if (imesa->Registers.TexAddr[1].ui != imesa->Registers.TexAddr[0].ui)
                {
                    imesa->Registers.TexAddr[1].ui = imesa->Registers.TexAddr[0].ui;
                    imesa->Registers.changed.ni.fTex1AddrChanged = GL_TRUE;
                }

                if (imesa->Registers.TexBlendCtrl[1].ui != TBC_Blend1)
                {
                    imesa->Registers.TexBlendCtrl[1].ui = TBC_Blend1;
                    imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
                }

                TexCtrl.ni.clrArg1Invert = GL_TRUE;
                imesa->bTexEn1 = GL_TRUE;
                break;

            case GL_LUMINANCE_ALPHA:
            case GL_RGBA:
                TexBlendCtrl.ui = TBC_BlendAlpha0;
                TexDescr.ni.tex1En = GL_TRUE;
                TexDescr.ni.texBLoopEn = GL_TRUE;
                TexDescr.ni.tex1Width  = TexDescr.ni.tex0Width;
                TexDescr.ni.tex1Height = TexDescr.ni.tex0Height;
                TexDescr.ni.tex1Fmt = TexDescr.ni.tex0Fmt;

                if (imesa->Registers.TexAddr[1].ui != imesa->Registers.TexAddr[0].ui)
                {
                    imesa->Registers.TexAddr[1].ui = imesa->Registers.TexAddr[0].ui;
                    imesa->Registers.changed.ni.fTex1AddrChanged = GL_TRUE;
                }

                if (imesa->Registers.TexBlendCtrl[1].ui != TBC_BlendAlpha1)
                {
                    imesa->Registers.TexBlendCtrl[1].ui = TBC_BlendAlpha1;
                    imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
                }

                TexCtrl.ni.clrArg1Invert = GL_TRUE;
                imesa->bTexEn1 = GL_TRUE;
                break;

            case GL_INTENSITY:
                TexBlendCtrl.ui = TBC_BlendInt0;
                TexDescr.ni.tex1En = GL_TRUE;
                TexDescr.ni.texBLoopEn = GL_TRUE;
                TexDescr.ni.tex1Width  = TexDescr.ni.tex0Width;
                TexDescr.ni.tex1Height = TexDescr.ni.tex0Height;
                TexDescr.ni.tex1Fmt = TexDescr.ni.tex0Fmt;

                if (imesa->Registers.TexAddr[1].ui != imesa->Registers.TexAddr[0].ui)
                {
                    imesa->Registers.TexAddr[1].ui = imesa->Registers.TexAddr[0].ui;
                    imesa->Registers.changed.ni.fTex1AddrChanged = GL_TRUE;
                }

                if (imesa->Registers.TexBlendCtrl[1].ui != TBC_BlendInt1)
                {
                    imesa->Registers.TexBlendCtrl[1].ui = TBC_BlendInt1;
                    imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
                }
                TexCtrl.ni.clrArg1Invert = GL_TRUE;
                TexCtrl.ni.alphaArg1Invert = GL_TRUE;
                imesa->bTexEn1 = GL_TRUE;
                break;
        }
        __HWEnvCombineSingleUnitScale(imesa, 0, 0, &TexBlendCtrl);
        break;

        /*
         GL_ADD
        */
    case GL_ADD:
        printf("Add\n");
        TexCtrl.ni.clrArg1Invert = GL_FALSE;
        TexBlendCtrl.ui = TBC_AddAlpha;
        __HWEnvCombineSingleUnitScale(imesa, 0, 0, &TexBlendCtrl);
        break;

#if GL_ARB_texture_env_combine
    case GL_COMBINE_ARB:
        __HWParseTexEnvCombine(imesa, 0, &TexCtrl, &TexBlendCtrl);
        break;
#endif

   default:
      fprintf(stderr, "unknown tex env mode");
      exit(1);
      break;			
   }

    TexCtrl.ni.uMode = !(t->texParams.sWrapMode & 0x01);
    TexCtrl.ni.vMode = !(t->texParams.tWrapMode & 0x01);

    switch (t->texParams.minFilter)
    {
        case GL_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Point;
            TexCtrl.ni.mipmapEnable = GL_FALSE;
            break;

        case GL_LINEAR:
            TexCtrl.ni.filterMode   = TFM_Bilin;
            TexCtrl.ni.mipmapEnable = GL_FALSE;
            break;

        case GL_NEAREST_MIPMAP_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Point;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;

        case GL_LINEAR_MIPMAP_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Bilin;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;

        case GL_NEAREST_MIPMAP_LINEAR:
        case GL_LINEAR_MIPMAP_LINEAR:
            TexCtrl.ni.filterMode   = TFM_Trilin;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;
    }

    if((ctx->Texture.Unit[0].LodBias !=0.0F) && (TexCtrl.ni.dBias != 0))
    {
    	GLfloat bias;
    	GLuint  ul;
    	GLint *pl;
    	
    	bias = ctx->Texture.Unit[0].LodBias;
    	pl   = (GLint *)&bias;
    	
    	/* if the value is >= 15.9375 determine whether >= 16
    	   or <0
    	*/
    	if(((*pl) & 0x7FFFFFFF) >= 0x417F0000)
    	{
    	    if((*pl) & 0x80000000)
    	    {
    	        ul=0x101;
    	    }
    	    else
    	    {
    	        ul=0xff;
    	    }
    	}
    	else
    	{
            ul=(GLuint)(bias*16.0);
        }
        
        ul &= 0x1FF;
        TexCtrl.ni.dBias = ul;
    }

    TexDescr.ni.tex0En = GL_TRUE;
    TexDescr.ni.tex0Width  = t->image[0].image->WidthLog2;
    TexDescr.ni.tex0Height = t->image[0].image->HeightLog2;
    TexDescr.ni.tex0Fmt = t->image[0].internalFormat;
    TexCtrl.ni.dMax = t->max_level;

    if (TexDescr.ni.tex1En)
        TexDescr.ni.texBLoopEn = GL_TRUE;

    if (imesa->Registers.TexAddr[0].ui != (GLuint)t->texParams.hwPhysAddress)
    {
        imesa->Registers.TexAddr[0].ui = (GLuint) t->texParams.hwPhysAddress | 0x2;
        
        if(t->heap == SAVAGE_AGP_HEAP)
            imesa->Registers.TexAddr[0].ui |= 0x1;
            
        imesa->Registers.changed.ni.fTex0AddrChanged = GL_TRUE;
    }

    if (imesa->Registers.TexCtrl[0].ui != TexCtrl.ui)
    {
        imesa->Registers.TexCtrl[0].ui = TexCtrl.ui;
        imesa->Registers.changed.ni.fTex0CtrlChanged = GL_TRUE;
    }

    if (imesa->Registers.TexBlendCtrl[0].ui != TexBlendCtrl.ui)
    {
        imesa->Registers.TexBlendCtrl[0].ui = TexBlendCtrl.ui;
        imesa->Registers.changed.ni.fTex0BlendCtrlChanged = GL_TRUE;
    }

    if (imesa->Registers.TexDescr.ui != TexDescr.ui)
    {
        imesa->Registers.TexDescr.ui = TexDescr.ui;
        imesa->Registers.changed.ni.fTexDescrChanged = GL_TRUE;
    }

   return;
}



static void savageUpdateTex1State( GLcontext *ctx )
{
   savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
   struct gl_texture_object	*tObj;
   savageTextureObjectPtr t;
   GLuint format;
   Reg_TexCtrl           TexCtrl;
   Reg_TexBlendCtrl      TexBlendCtrl;
   Reg_TexDescr          TexDescr;


   /* disable */
   if(imesa->bTexEn1)
   {
       imesa->bTexEn1 = GL_FALSE;
       return;
   }

   if (ctx->Texture.Unit[1].ReallyEnabled == 0) {
      imesa->Registers.TexDescr.ni.tex1En = GL_FALSE;
      imesa->Registers.TexBlendCtrl[1].ui = TBC_NoTexMap1;
      imesa->Registers.TexCtrl[1].ui = 0x20f040;
      imesa->Registers.TexAddr[1].ui = 0;
      imesa->Registers.TexDescr.ni.texBLoopEn = GL_FALSE;
      imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTexDescrChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
      imesa->Registers.changed.ni.fTex1AddrChanged = GL_TRUE;
      return;
   }

   tObj = ctx->Texture.Unit[1].Current;

   if (ctx->Texture.Unit[1].ReallyEnabled != TEXTURE0_2D ||
       tObj->Image[tObj->BaseLevel]->Border > 0) {
      /* 1D or 3D texturing enabled, or texture border - fallback */
      imesa->Fallback |= SAVAGE_FALLBACK_TEXTURE;
      return;
   }

   /* Do 2D texture setup */

   t = tObj->DriverData;
   if (!t) {
      t = savageCreateTexObj( imesa, tObj );
      if (!t)
         return;
   }
    
   if (t->current_unit != 1)
      savageTexSetUnit( t, 1 );

   imesa->CurrentTexObj[1] = t;

   t->bound = 2;

   if (t->dirty_images)
       savageUploadTexImages(imesa, imesa->CurrentTexObj[1]);
   
   if (t->MemBlock)
      savageUpdateTexLRU( imesa, t );

   TexDescr.ui = imesa->Registers.TexDescr.ui;
   TexCtrl.ui = imesa->Registers.TexCtrl[1].ui;
   TexBlendCtrl.ui = imesa->Registers.TexBlendCtrl[1].ui;

   format = t->image[0].baseFormat;
   TexDescr.ni.tex1Fmt = t->image[0].internalFormat;


   switch (ctx->Texture.Unit[1].EnvMode) {
   case GL_REPLACE:
        TexCtrl.ni.clrArg1Invert = GL_FALSE;
        switch (format)
        {
            case GL_LUMINANCE:
            case GL_RGB:
                TexBlendCtrl.ui = TBC_Decal;
                break;

            case GL_LUMINANCE_ALPHA:
            case GL_INTENSITY:
            case GL_RGBA:
                TexBlendCtrl.ui = TBC_Copy;
                break;

            case GL_ALPHA:
                TexBlendCtrl.ui = TBC_CopyAlpha1;
                break;
        }
        __HWEnvCombineSingleUnitScale(imesa, 0, 1, &TexBlendCtrl);
      break;
   case GL_MODULATE:
       TexCtrl.ni.clrArg1Invert = GL_FALSE;
       TexBlendCtrl.ui = TBC_ModulAlpha1;
       __HWEnvCombineSingleUnitScale(imesa, 0, 1, &TexBlendCtrl);
       break;

/*#if GL_EXT_texture_env_add*/
    case GL_ADD:
        TexCtrl.ni.clrArg1Invert = GL_FALSE;
        TexBlendCtrl.ui = TBC_AddAlpha1;
        __HWEnvCombineSingleUnitScale(imesa, 0, 1, &TexBlendCtrl);
        break;
/*#endif*/

#if GL_ARB_texture_env_combine
    case GL_COMBINE_ARB:
        __HWParseTexEnvCombine(imesa, 1, &TexCtrl, &TexBlendCtrl);
        break;
#endif

   case GL_DECAL:
        TexCtrl.ni.clrArg1Invert = GL_FALSE;

        switch (format)
        {
            case GL_LUMINANCE:
            case GL_RGB:
                TexBlendCtrl.ui = TBC_Decal1;
                break;
            case GL_LUMINANCE_ALPHA:
            case GL_INTENSITY:
            case GL_RGBA:
                TexBlendCtrl.ui = TBC_DecalAlpha1;
                break;

                /*
                // GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_ALPHA, GL_INTENSITY
                // are undefined with GL_DECAL
                */
            case GL_ALPHA:
                TexBlendCtrl.ui = TBC_CopyAlpha1;
                break;
        }
        __HWEnvCombineSingleUnitScale(imesa, 0, 1, &TexBlendCtrl);
        break;

   case GL_BLEND:
        if (format == GL_LUMINANCE)
        {
            /*
            // This is a hack for GLQuake, invert.
            */
            TexCtrl.ni.clrArg1Invert = GL_TRUE;
            TexBlendCtrl.ui = 0;
        }
        __HWEnvCombineSingleUnitScale(imesa, 0, 1, &TexBlendCtrl);
      break;

   default:
      fprintf(stderr, "unkown tex 1 env mode\n");
      exit(1);
      break;			
   }

    TexCtrl.ni.uMode = !(t->texParams.sWrapMode & 0x01);
    TexCtrl.ni.vMode = !(t->texParams.tWrapMode & 0x01);

    switch (t->texParams.minFilter)
    {
        case GL_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Point;
            TexCtrl.ni.mipmapEnable = GL_FALSE;
            break;

        case GL_LINEAR:
            TexCtrl.ni.filterMode   = TFM_Bilin;
            TexCtrl.ni.mipmapEnable = GL_FALSE;
            break;

        case GL_NEAREST_MIPMAP_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Point;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;

        case GL_LINEAR_MIPMAP_NEAREST:
            TexCtrl.ni.filterMode   = TFM_Bilin;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;

        case GL_NEAREST_MIPMAP_LINEAR:
        case GL_LINEAR_MIPMAP_LINEAR:
            TexCtrl.ni.filterMode   = TFM_Trilin;
            TexCtrl.ni.mipmapEnable = GL_TRUE;
            break;
    }
    
    if((ctx->Texture.Unit[1].LodBias !=0.0F)&&(TexCtrl.ni.dBias != 0))
    {
    	GLfloat bias;
    	GLuint  ul;
    	GLint *pl;
    	
    	bias = ctx->Texture.Unit[1].LodBias;
    	pl   = (GLint *)&bias;
    	
    	/* if the value is >= 15.9375 determine whether >= 16
    	   or <0
    	*/
    	if(((*pl) & 0x7FFFFFFF) >= 0x417F0000)
    	{
    	    if((*pl) & 0x80000000)
    	    {
    	        ul=0x101;
    	    }
    	    else
    	    {
    	        ul=0xff;
    	    }
    	}
    	else
    	{
            ul=(GLuint)(bias*16.0);
        }
        
        ul &= 0x1FF;
        TexCtrl.ni.dBias = ul;
    }

    TexDescr.ni.tex1En = GL_TRUE;
    TexDescr.ni.tex1Width  = t->image[0].image->WidthLog2;
    TexDescr.ni.tex1Height = t->image[0].image->HeightLog2;
    TexDescr.ni.tex1Fmt = t->image[0].internalFormat;
    TexCtrl.ni.dMax = t->max_level;
    TexDescr.ni.texBLoopEn = GL_TRUE;

    if (imesa->Registers.TexAddr[1].ui != (GLuint)t->texParams.hwPhysAddress)
    {
        imesa->Registers.TexAddr[1].ui = (GLuint) t->texParams.hwPhysAddress| 2;
        
        if(t->heap == SAVAGE_AGP_HEAP)
           imesa->Registers.TexAddr[1].ui |= 0x1;
        
        /*imesa->Registers.TexAddr[1].ui = (GLuint) t->texParams.hwPhysAddress| 3;*/
        imesa->Registers.changed.ni.fTex1AddrChanged = GL_TRUE;
    }

    if (imesa->Registers.TexCtrl[1].ui != TexCtrl.ui)
    {
        imesa->Registers.TexCtrl[1].ui = TexCtrl.ui;
        imesa->Registers.changed.ni.fTex1CtrlChanged = GL_TRUE;
    }

    if (imesa->Registers.TexBlendCtrl[1].ui != TexBlendCtrl.ui)
    {
        imesa->Registers.TexBlendCtrl[1].ui = TexBlendCtrl.ui;
        imesa->Registers.changed.ni.fTex1BlendCtrlChanged = GL_TRUE;
    }

    if (imesa->Registers.TexDescr.ui != TexDescr.ui)
    {
        imesa->Registers.TexDescr.ui = TexDescr.ui;
        imesa->Registers.changed.ni.fTexDescrChanged = GL_TRUE;
    }

}


void savageUpdateTextureState( GLcontext *ctx )
{
   savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
   if (imesa->CurrentTexObj[0]) imesa->CurrentTexObj[0]->bound = 0;
   if (imesa->CurrentTexObj[1]) imesa->CurrentTexObj[1]->bound = 0;
   imesa->CurrentTexObj[0] = 0;
   imesa->CurrentTexObj[1] = 0;   
   imesa->Fallback &= ~SAVAGE_FALLBACK_TEXTURE;
   savageUpdateTex0State( ctx );
   savageUpdateTex1State( ctx );
   SAVAGE_CONTEXT( ctx )->dirty |= (SAVAGE_UPLOAD_CTX |
                                  SAVAGE_UPLOAD_TEX0 | 
                                  SAVAGE_UPLOAD_TEX1);
}



/*****************************************
 * DRIVER functions
 *****************************************/

static void savageTexEnv( GLcontext *ctx, GLenum target, 
			GLenum pname, const GLfloat *param )
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );

   if (pname == GL_TEXTURE_ENV_MODE) {

      imesa->new_state |= SAVAGE_NEW_TEXTURE;

   } else if (pname == GL_TEXTURE_ENV_COLOR) {

      struct gl_texture_unit *texUnit = 
	 &ctx->Texture.Unit[ctx->Texture.CurrentUnit];
      const GLfloat *fc = texUnit->EnvColor;
      GLuint r, g, b, a, col;
      FLOAT_COLOR_TO_UBYTE_COLOR(r, fc[0]);
      FLOAT_COLOR_TO_UBYTE_COLOR(g, fc[1]);
      FLOAT_COLOR_TO_UBYTE_COLOR(b, fc[2]);
      FLOAT_COLOR_TO_UBYTE_COLOR(a, fc[3]);

      col = ((a << 24) | 
	     (r << 16) | 
	     (g <<  8) | 
	     (b <<  0));
    

   } 
}

static void savageTexImage( GLcontext *ctx, 
			  GLenum target,
			  struct gl_texture_object *tObj, 
			  GLint level,
			  GLint internalFormat,
			  const struct gl_texture_image *image )
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
   savageTextureObjectPtr t;


   if (target != GL_TEXTURE_2D)
      return;

   if (level >= SAVAGE_TEX_MAXLEVELS)
      return;

   t = (savageTextureObjectPtr) tObj->DriverData;
   if (t) {
      /* if this is the current object, it will force an update */

      savageDestroyTexObj( imesa, t );
      tObj->DriverData = 0;
      imesa->new_state |= SAVAGE_NEW_TEXTURE;
   }
}

static void savageTexSubImage( GLcontext *ctx, GLenum target,
			     struct gl_texture_object *tObj, GLint level,
			     GLint xoffset, GLint yoffset,
			     GLsizei width, GLsizei height,
			     GLint internalFormat,
			     const struct gl_texture_image *image ) 
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
   savageTextureObjectPtr t;

   if ( target != GL_TEXTURE_2D ) 
      return;
   
   t = (savageTextureObjectPtr) tObj->DriverData;
   if (t) {

      savageDestroyTexObj( imesa, t );
      tObj->DriverData = 0;
      imesa->new_state |= SAVAGE_NEW_TEXTURE;
   }
}

static void savageTexParameter( GLcontext *ctx, GLenum target,
			      struct gl_texture_object *tObj,
			      GLenum pname, const GLfloat *params )
{
   savageTextureObjectPtr t = (savageTextureObjectPtr) tObj->DriverData;
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );

   if (!t || target != GL_TEXTURE_2D)
      return;

   switch (pname) {
   case GL_TEXTURE_MIN_FILTER:
   case GL_TEXTURE_MAG_FILTER:
      savageSetTexFilter(imesa, t,tObj->MinFilter,tObj->MagFilter);
      break;

   case GL_TEXTURE_WRAP_S:
   case GL_TEXTURE_WRAP_T:
      savageSetTexWrapping(t,tObj->WrapS,tObj->WrapT);
      break;
  
   case GL_TEXTURE_BORDER_COLOR:
      savageSetTexBorderColor(t,tObj->BorderColor);
      break;

   default:
      return;
   }

   imesa->new_state |= SAVAGE_NEW_TEXTURE;
}

static void savageBindTexture( GLcontext *ctx, GLenum target,
			     struct gl_texture_object *tObj )
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
   

   if (imesa->CurrentTexObj[ctx->Texture.CurrentUnit]) {
      imesa->CurrentTexObj[ctx->Texture.CurrentUnit]->bound = 0;
      imesa->CurrentTexObj[ctx->Texture.CurrentUnit] = 0;  
   }

   imesa->new_state |= SAVAGE_NEW_TEXTURE;
}

static void savageDeleteTexture( GLcontext *ctx, struct gl_texture_object *tObj )
{
   savageTextureObjectPtr t = (savageTextureObjectPtr)tObj->DriverData;
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );

   if (t) {

      if (t->bound) {
	 imesa->CurrentTexObj[t->bound-1] = 0;
	 imesa->new_state |= SAVAGE_NEW_TEXTURE;
      }

      savageDestroyTexObj(imesa,t);
      tObj->DriverData=0;
   }
}


static GLboolean savageIsTextureResident( GLcontext *ctx, 
					struct gl_texture_object *t )
{
   savageTextureObjectPtr mt;

/*     LOCK_HARDWARE; */
   mt = (savageTextureObjectPtr)t->DriverData;
/*     UNLOCK_HARDWARE; */

   return mt && mt->MemBlock;
}

void savageDDInitTextureFuncs( GLcontext *ctx )
{
   ctx->Driver.TexEnv = savageTexEnv;
   ctx->Driver.TexImage = savageTexImage;
   ctx->Driver.TexSubImage = savageTexSubImage;
   ctx->Driver.BindTexture = savageBindTexture;
   ctx->Driver.DeleteTexture = savageDeleteTexture;
   ctx->Driver.TexParameter = savageTexParameter;
   ctx->Driver.UpdateTexturePalette = 0;
   ctx->Driver.IsTextureResident = savageIsTextureResident;
}
