/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#include <stdio.h>
#include <math.h>

#include "types.h"
#include "vb.h"
#include "pipeline.h"

#include "mm.h"
#include "savagetris.h"
#include "savagevb.h"
#include "savagelog.h"
#include "savagestate.h"
#include "pb.h"

extern void gl_reduced_prim_change(GLcontext *ctx,GLenum prim);
/* Used in savagetritmp.h
 */
#define SAVAGE_COLOR(to, from) {			\
  (to)[0] = (from)[2];				\
  (to)[1] = (from)[1];				\
  (to)[2] = (from)[0];				\
  (to)[3] = (from)[3];				\
}

static triangle_func tri_tab[0x10];   
static quad_func     quad_tab[0x10];  
static line_func     line_tab[0x10];  
static points_func   points_tab[0x10];

#define IND (0)
#define TAG(x) x
#include "savagetritmp.h"

#define IND (SAVAGE_OFFSET_BIT)	
#define TAG(x) x##_offset
#include "savagetritmp.h"

#define IND (SAVAGE_TWOSIDE_BIT)	
#define TAG(x) x##_twoside
#include "savagetritmp.h"

#define IND (SAVAGE_TWOSIDE_BIT|SAVAGE_OFFSET_BIT) 
#define TAG(x) x##_twoside_offset
#include "savagetritmp.h"

/* Create the raw tab which include the drawing strip*/
#define FLUSH_PRIM(prim)			\
do {						\
   if (ctx->PB->primitive != prim) {		\
     gl_reduced_prim_change( ctx, prim );	\
     }						\
} while(0)

#define FLUSH_POLY(prim)					\
do {								\
   if ((ctx->IndirectTriangles & DD_TRI_UNFILLED) == 0) {	\
      FLUSH_PRIM(prim);						\
   }								\
} while(0)

#define RENDER_POINTS( start, count ) \
   (*ctx->Driver.PointsFunc)( ctx, start, count )

#define RENDER_LINE( i1, i ) \
   (*ctx->Driver.LineFunc)( ctx, i1, i, i )

#define RENDER_TRI( i2, i1, i, pv, parity )	\
do {						\
   if (parity)					\
      ctx->TriangleFunc( ctx, i1, i2, i, pv );	\
   else						\
      ctx->TriangleFunc( ctx, i2, i1, i, pv );	\
} while (0)


#define RENDER_QUAD( i3, i2, i1, i, pv )	\
   ctx->QuadFunc( ctx, i3, i2, i1, i, i );

#define TAG(x) savage_##x##_raw

#define LOCAL_VARS  				\
    GLcontext *ctx = VB->ctx;			\
    GLubyte *eflag = VB->EdgeFlagPtr->data;	\
    GLuint *stipplecounter = &VB->ctx->StippleCounter; \
    (void) eflag; (void) stipplecounter;

#define INIT(x)  if (x == GL_POLYGON) FLUSH_POLY(x); else FLUSH_PRIM(x)
#define RESET_STIPPLE *stipplecounter = 0
#include "savagerendertmp.h"


void savageDDTrifuncInit()
{
   init();
   init_offset();
   init_twoside();
   init_twoside_offset();
}


#define ALL_FALLBACK (DD_SELECT | DD_FEEDBACK )
#define POINT_FALLBACK (ALL_FALLBACK)
#define LINE_FALLBACK (ALL_FALLBACK | DD_LINE_STIPPLE|DD_LINE_SMOOTH)

#define TRI_FALLBACK (ALL_FALLBACK | DD_TRI_UNFILLED|DD_TRI_SMOOTH)

#define ANY_FALLBACK (POINT_FALLBACK|LINE_FALLBACK|TRI_FALLBACK|DD_TRI_STIPPLE)
#define ANY_RASTER_FLAGS (DD_TRI_LIGHT_TWOSIDE|DD_TRI_OFFSET)

GLboolean savageMultipassFunc(struct vertex_buffer *VB,GLuint count)
{
    GLcontext *ctx = VB->ctx;
    savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
    __DRIdrawablePrivate *dPriv = imesa->driDrawable;
    XF86DRIClipRectPtr pbox;
    
    imesa->currentClip--;
    if (imesa->currentClip <= 0)
        return 0;/* need not render again*/
    pbox = dPriv->pClipRects;
    if (!pbox)
        return 0;
    pbox += (imesa->currentClip-1);
    
    /* set scissor to the first clip box*/
    savageDDScissor(ctx,pbox->x1,pbox->y1,pbox->x2,pbox->y2);
    UNLOCK_HARDWARE(imesa);
    savageDDUpdateHwState(ctx); /* update to hardware register*/
    LOCK_HARDWARE(imesa);
    return 1;/*re-render*/
}

void savageDDChooseRenderState(GLcontext *ctx)
{
    savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
    GLuint         flags   = ctx->TriangleCaps;
    CARD32 index    = 0;

    if (imesa->Fallback) {
        
        imesa->renderindex = SAVAGE_FALLBACK_BIT;
	return;
    }
    
    if (flags & ANY_RASTER_FLAGS) {
	if (flags & DD_TRI_LIGHT_TWOSIDE)       index |= SAVAGE_TWOSIDE_BIT;
	if (flags & DD_TRI_OFFSET)              index |= SAVAGE_OFFSET_BIT; 
    }

    /*savage_render_vb_raw*/
    if ((!(flags & ANY_FALLBACK))&&(!(flags & DD_TRI_LIGHT_TWOSIDE)) &&
                                    (!(flags & DD_TRI_OFFSET)))
    {
       ctx->Driver.RenderVBRawTab = savage_render_tab_raw;
    }
    imesa->PointsFunc = points_tab[index];
    imesa->LineFunc = line_tab[index];
    imesa->TriangleFunc = tri_tab[index];
    imesa->QuadFunc = quad_tab[index];
    imesa->renderindex = index;
    imesa->IndirectTriangles = 0;

    if (flags & ANY_FALLBACK) {
	if (flags & POINT_FALLBACK) {
	    imesa->renderindex |= SAVAGE_FALLBACK_BIT;
	    imesa->PointsFunc = 0;
	    imesa->IndirectTriangles |= DD_POINT_SW_RASTERIZE;
	}
	    
	if (flags & LINE_FALLBACK) {
	    imesa->renderindex |= SAVAGE_FALLBACK_BIT;
	    imesa->LineFunc = 0;
	    imesa->IndirectTriangles |= DD_LINE_SW_RASTERIZE;
	}

	if (flags & TRI_FALLBACK) {
	    imesa->renderindex |= SAVAGE_FALLBACK_BIT;
	    imesa->TriangleFunc = 0;
	    imesa->QuadFunc = 0;
	    imesa->IndirectTriangles |= (DD_TRI_SW_RASTERIZE |
					 DD_QUAD_SW_RASTERIZE);
	}
	/* Special cases:
	 */
	if ((flags & DD_TRI_STIPPLE) &&
	    (ctx->IndirectTriangles & DD_TRI_STIPPLE)) {
	    imesa->renderindex |= SAVAGE_FALLBACK_BIT;
	    imesa->TriangleFunc = 0;
	    imesa->QuadFunc = 0;
	    imesa->IndirectTriangles |= (DD_TRI_SW_RASTERIZE |
					 DD_QUAD_SW_RASTERIZE);
	}
    }
}



