/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

 
#ifndef SAVAGETRIS_INC 
#define SAVAGETRIS_INC 
 
#include "types.h" 
#include "savageioctl.h" 
#include "savagevb.h" 
#include "savagecontext.h" 
#include "savage_bci.h" 
 
extern void savagePrintRenderState( const char *msg, GLuint state ); 
extern void savageDDChooseRenderState(GLcontext *ctx); 
extern void savageDDTrifuncInit( void ); 
GLboolean savageMultipassFunc(struct vertex_buffer *,GLuint); 
 
#define SAVAGE_OFFSET_BIT	     0x1
#define SAVAGE_TWOSIDE_BIT       0x2 
#define SAVAGE_FALLBACK_BIT      0x8 
 
#define SAVAGE_Z_OR_W_BIT 	     0x10 
#define SAVAGE_COLOR_BIT	     0x10 
#define SAVAGE_SPECULAR_BIT      0x20 
#define SAVAGE_T0_BIT            0x40 
#define SAVAGE_T1_BIT            0x80 
#define SAVAGE_T0_SIZE           0x100 
#define SAVAGE_T1_SIZE           0x200 
 
#define LINE_AND_PT_OFFSET       0.5f 
 
 
static  __inline__ volatile GLuint * savage_send_one_vertex(savageContextPtr imesa, savageVertexPtr v, volatile GLuint * vb, GLuint extra, GLuint start) 
{ 
    GLuint j; 
    for (j = start ; j < 5 ; j++) 
    { 
        WRITE_CMD(vb, v->ui[j],GLuint); 
    }
        
    if(extra) 
    { 
       if(!(imesa->DrawPrimitiveCmd & SAVAGE_HW_NO_CS)) 
       { 
           WRITE_CMD(vb, v->ui[5],GLuint); 
       } 
          
       if(!(imesa->DrawPrimitiveCmd & SAVAGE_HW_NO_UV0)) 
       { 
           WRITE_CMD(vb, v->ui[6],GLuint); 
           WRITE_CMD(vb, v->ui[7],GLuint); 
       } 
       
       if(!(imesa->DrawPrimitiveCmd & SAVAGE_HW_NO_UV1)) 
       { 
           WRITE_CMD(vb, v->ui[8],GLuint); 
           WRITE_CMD(vb, v->ui[9],GLuint); 
       } 
    }
    return vb; 
} 
 
static void __inline__ savage_draw_triangle( savageContextPtr imesa, 
					   savageVertexPtr v0, 
					   savageVertexPtr v1, 
					   savageVertexPtr v2 ) 
{ 
   GLuint vertsize = imesa->vertsize; 
   GLuint extra; 
#if SAVAGEDEBUG
   volatile GLuint *vb = SAVAGE_GET_WRITE_POINTER1(imesa, 3 * vertsize + 1 +8); 
#else
   volatile GLuint *vb = SAVAGE_GET_WRITE_POINTER1(imesa, 4 * vertsize + 1); 
#endif

   imesa->DrawPrimitiveCmd &=
       ~(SAVAGE_HW_TRIANGLE_TYPE| SAVAGE_HW_TRIANGLE_CONT);
   WRITE_CMD(vb,SAVAGE_DRAW_PRIMITIVE(3, imesa->DrawPrimitiveCmd, 0),GLuint);

   extra = (vertsize > 5) ? 1:0; 
 
   vb = savage_send_one_vertex(imesa, v0, vb, extra, 0); 
   vb = savage_send_one_vertex(imesa, v1, vb, extra, 0); 
   vb = savage_send_one_vertex(imesa, v2, vb, extra, 0); 
 
#if SAVAGEDEBUG
   {
        GLuint x0,y0,w,h;
        x0 = (GLuint)imesa->drawX;
        y0 = (GLuint)imesa->drawY;
        w  = (GLuint)imesa->driDrawable->w;
        h  = (GLuint)imesa->driDrawable->h;

   	(*vb) = 0x4BCC00C0;
   	vb++;
   	(*vb) = imesa->savageScreen->backOffset;
   	vb++;
   	(*vb) = imesa->savageScreen->backBitmapDesc;
   	vb++;
   	(*vb) = (y0<<16)|x0;
   	vb++;
   	(*vb) = 0x0;
   	vb++;
   	(*vb) = (h<<16)|w;
   	vb++;
   }
#endif
   imesa->DMABuf.writePointer = (GLuint)vb;
   SAVAGE_WRITTEN_ENTRIES(imesa, 3 * imesa->vertsize + 1); 
} 
 
static __inline__ void savage_draw_triangle_strip(
    struct vertex_buffer *VB,
    GLuint start,
    GLuint count)
{
    GLcontext *ctx = VB->ctx;
    savageVertexPtr ivert = SAVAGE_DRIVER_DATA(VB)->verts;
    savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
    GLuint vertsize = imesa->vertsize;
    volatile GLuint *vb; 
    int i,end;
    GLuint extra;
    
    imesa->DrawPrimitiveCmd &=
        ~(SAVAGE_HW_TRIANGLE_TYPE | SAVAGE_HW_TRIANGLE_CONT);
    imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_STRIP;
    end = count;
    count=count - start;
    extra = (vertsize > 5) ? 1:0;
    
    while( count>0 )
    {
        if (count >255)
            count = 255;
        
#if (SAVAGE_CMD_DMA)
        if ((count * vertsize +1) >= DMA_PAGE_SIZE/4)
            count = (DMA_PAGE_SIZE/4 - 1)/vertsize ;
#endif
        
#if SAVAGEDEBUG
        vb = SAVAGE_GET_WRITE_POINTER1(imesa, count* vertsize + 1 +8); 
#else
        vb = SAVAGE_GET_WRITE_POINTER1(imesa, count * vertsize + 1); 
#endif         
        WRITE_CMD(vb,
                  SAVAGE_DRAW_PRIMITIVE(count, imesa->DrawPrimitiveCmd, 0),GLuint);
        for (i=start;i<start+count;i++)
        {
            vb = savage_send_one_vertex(imesa, &ivert[i], vb, extra, 0);     
        }

        start = start+count;
        count = end - start;

        imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_CONT;
    }
    
    imesa->DMABuf.writePointer = (GLuint)vb;
    SAVAGE_WRITTEN_ENTRIES(imesa, 3 * imesa->vertsize + 1);
    
}    

static __inline__ void savage_draw_triangle_fan(
    struct vertex_buffer *VB,
    GLuint start,
    GLuint count)
{
    GLcontext *ctx = VB->ctx;
    savageVertexPtr ivert = SAVAGE_DRIVER_DATA(VB)->verts;
    savageContextPtr imesa = SAVAGE_CONTEXT(ctx);
    GLuint vertsize = imesa->vertsize;
    volatile GLuint *vb; 
    int i,end;
    GLuint extra;
    
    imesa->DrawPrimitiveCmd &=
        ~(SAVAGE_HW_TRIANGLE_TYPE | SAVAGE_HW_TRIANGLE_CONT);
    imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_FAN;
    end = count;
    count=count - start;
    extra = (vertsize > 5) ? 1:0;
    
    while( count>0 )
    {
        if (count >255)
            count = 255;
        
#if (SAVAGE_CMD_DMA)
        if ((count * vertsize +1) >= DMA_PAGE_SIZE/4)
            count = (DMA_PAGE_SIZE/4 - 1)/vertsize ;
#endif
     
#if SAVAGEDEBUG
        vb = SAVAGE_GET_WRITE_POINTER1(imesa, count* vertsize + 1 +8); 
#else
        vb = SAVAGE_GET_WRITE_POINTER1(imesa, count * vertsize + 1); 
#endif         
        WRITE_CMD(vb,
                  SAVAGE_DRAW_PRIMITIVE(count, imesa->DrawPrimitiveCmd, 0),GLuint);
        for (i=start;i<start+count;i++)
        {
            vb = savage_send_one_vertex(imesa, &ivert[i], vb, extra, 0);     
        }

        start = start+count;
        count = end - start;
        imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_CONT;        
    }
    
    imesa->DMABuf.writePointer = (GLuint)vb;
    SAVAGE_WRITTEN_ENTRIES(imesa, 3 * imesa->vertsize + 1);
}    
 
static __inline__ void savage_draw_point( savageContextPtr imesa, 
					savageVertexPtr tmp, 
					float sz ) 
{ 
   int vertsize = imesa->vertsize; 
   volatile GLuint *vb = SAVAGE_GET_WRITE_POINTER1(imesa, 4 * vertsize + 1); 
   const GLfloat x = tmp->v.x; 
   const GLfloat y = tmp->v.y; 
   GLuint extra;
   
   imesa->DrawPrimitiveCmd &=
       ~(SAVAGE_HW_TRIANGLE_TYPE | SAVAGE_HW_TRIANGLE_CONT);   
   imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_FAN; 
     
   WRITE_CMD(vb, SAVAGE_DRAW_PRIMITIVE(4, imesa->DrawPrimitiveCmd, 0),GLuint); 

   extra = (vertsize > 5) ? 1:0;
 
   WRITE_CMD(vb, x - sz, GLfloat);
   WRITE_CMD(vb, y - sz, GLfloat);

   vb = savage_send_one_vertex(imesa, tmp, vb, extra, 2); 
 
 
   WRITE_CMD(vb, x + sz, GLfloat);
   WRITE_CMD(vb, y - sz, GLfloat);

   vb = savage_send_one_vertex(imesa, tmp, vb, extra, 2); 
   WRITE_CMD(vb, x + sz, GLfloat);
   WRITE_CMD(vb, y + sz, GLfloat);

   vb = savage_send_one_vertex(imesa, tmp, vb, extra, 2); 
 
   WRITE_CMD(vb, x - sz, GLfloat);
   WRITE_CMD(vb, y + sz, GLfloat);

   vb = savage_send_one_vertex(imesa, tmp, vb, extra, 2); 
   imesa->DMABuf.writePointer = (GLuint)vb;    
   SAVAGE_WRITTEN_ENTRIES(imesa, 4 * imesa->vertsize + 1); 
} 
 
 
static __inline__ void savage_draw_line( savageContextPtr imesa, 
				       savageVertexPtr v0, 
				       savageVertexPtr v1 ) 
{  
  GLuint vertsize = imesa->vertsize; 
  volatile GLuint *vb = SAVAGE_GET_WRITE_POINTER1(imesa, 4 * vertsize + 1); 
   GLfloat dx, dy, sdx, sdy; 
   GLuint extra; 
   savageVertexPtr plowright; 
 
   extra = (vertsize > 5) ? 1:0;
   imesa->DrawPrimitiveCmd &=
       ~(SAVAGE_HW_TRIANGLE_TYPE | SAVAGE_HW_TRIANGLE_CONT);
   imesa->DrawPrimitiveCmd |= SAVAGE_HW_TRIANGLE_FAN; 
   WRITE_CMD(vb, SAVAGE_DRAW_PRIMITIVE(4, imesa->DrawPrimitiveCmd, 0),GLuint); 
   
   sdx = v1->v.x - v0->v.x; 
   sdy = v1->v.y - v0->v.y; 
   dx = fabs(sdx); 
   dy = fabs(sdy); 
 
   plowright = v1; 
   if((sdx < 0) || ((sdy == 0) && (sdy < 0))) 
   { 
       plowright = v0; 
       sdy = -sdy; 
   } 
 
   if(dx != 0) 
   { 
       if(dy > dx) 
       { 
           plowright->v.x += .5f * dx / dy; 
           if(sdy > 0) 
               plowright->v.y += 0.5f; 
           else 
               if(sdy < 0) 
                   plowright->v.y -= 0.5f; 
       } 
       else 
       { 
           plowright->v.x += 0.5f; 
           plowright->v.y -= 0.5f * sdy / dx; 
       } 
   } 
   else if(dy != 0) plowright->v.y += 0.5f; 
 
   if(dx > dy) 
   { 
       WRITE_CMD_FLOAT(vb, v0->v.x, GLfloat);
       WRITE_CMD_FLOAT(vb,( v0->v.y  - LINE_AND_PT_OFFSET), GLfloat);
 
       vb = savage_send_one_vertex(imesa, v0, vb, extra, 2); 

       WRITE_CMD_FLOAT(vb, v1->v.x, GLfloat);
       WRITE_CMD_FLOAT(vb,( v1->v.y  - LINE_AND_PT_OFFSET), GLfloat);     
 
       vb = savage_send_one_vertex(imesa, v1, vb, extra, 2); 
       WRITE_CMD_FLOAT(vb, v1->v.x, GLfloat);
       WRITE_CMD_FLOAT(vb,( v1->v.y  + LINE_AND_PT_OFFSET), GLfloat);
    
       vb = savage_send_one_vertex(imesa, v1, vb, extra, 2); 
       WRITE_CMD_FLOAT(vb, v0->v.x, GLfloat);
       WRITE_CMD_FLOAT(vb,( v0->v.y  + LINE_AND_PT_OFFSET), GLfloat);
   
       vb = savage_send_one_vertex(imesa, v0, vb, extra, 2); 
 
   } 
   else 
   { 
       WRITE_CMD_FLOAT(vb,( v0->v.x - LINE_AND_PT_OFFSET), GLfloat);
       WRITE_CMD_FLOAT(vb,v0->v.y, GLfloat);
   
       vb = savage_send_one_vertex(imesa, v0, vb, extra, 2); 
       WRITE_CMD_FLOAT(vb,( v1->v.x - LINE_AND_PT_OFFSET), GLfloat);
       WRITE_CMD_FLOAT(vb,v1->v.y, GLfloat);
 
       vb = savage_send_one_vertex(imesa, v1, vb, extra, 2); 
       WRITE_CMD_FLOAT(vb,( v1->v.x + LINE_AND_PT_OFFSET), GLfloat);
       WRITE_CMD_FLOAT(vb,v1->v.y, GLfloat);
       
       vb = savage_send_one_vertex(imesa, v1, vb, extra, 2); 
       WRITE_CMD_FLOAT(vb,( v0->v.x + LINE_AND_PT_OFFSET), GLfloat);
       WRITE_CMD_FLOAT(vb,v0->v.y, GLfloat);
     
       vb = savage_send_one_vertex(imesa, v0, vb, extra, 2); 
 
   } 
   imesa->DMABuf.writePointer = (GLuint)vb; 
   SAVAGE_WRITTEN_ENTRIES(imesa, 4 * imesa->vertsize + 1); 
} 
 
static void __inline__ savage_draw_quad( savageContextPtr imesa, 
					 savageVertexPtr v0, 
					 savageVertexPtr v1, 
					 savageVertexPtr v2, 
					 savageVertexPtr v3 ) 
{ 
   GLuint vertsize = imesa->vertsize; 
   volatile GLuint *vb = SAVAGE_GET_WRITE_POINTER1(imesa, 4 * vertsize + 1); 
   GLuint extra; 

   imesa->DrawPrimitiveCmd &=
       ~(SAVAGE_HW_TRIANGLE_TYPE | SAVAGE_HW_TRIANGLE_CONT);
   imesa->DrawPrimitiveCmd |= SAVAGE_HW_QUAD;
   WRITE_CMD(vb, SAVAGE_DRAW_PRIMITIVE(4, imesa->DrawPrimitiveCmd, 0),GLuint); 
 
   extra = (vertsize > 5) ? 1:0; 
   vb = savage_send_one_vertex(imesa, v0, vb, extra, 0); 
   vb = savage_send_one_vertex(imesa, v1, vb, extra, 0); 
   vb = savage_send_one_vertex(imesa, v2, vb, extra, 0); 
   vb = savage_send_one_vertex(imesa, v3, vb, extra, 0); 
   imesa->DMABuf.writePointer = (GLuint)vb; 
   SAVAGE_WRITTEN_ENTRIES(imesa, 4 * imesa->vertsize + 1); 
} 

#endif 
