/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA, S3 GRAPHICS, AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


static __inline void TAG(triangle)(GLcontext *ctx,
				   GLuint e0, GLuint e1, GLuint e2,
				   GLuint pv)
{
   savageContextPtr        savagectx   = SAVAGE_CONTEXT(ctx);
   struct vertex_buffer *VB        = ctx->VB;
   savageVertexPtr         savageverts = SAVAGE_DRIVER_DATA(VB)->verts;
   savageVertex     *v[3];

#if (IND & SAVAGE_OFFSET_BIT)
   GLfloat offset;
   GLfloat z[3];
#endif

#if (IND & SAVAGE_TWOSIDE_BIT )
   GLuint c[3];
#endif

   v[0] = &savageverts[e0];
   v[1] = &savageverts[e1];
   v[2] = &savageverts[e2];

#if (IND & SAVAGE_TWOSIDE_BIT)
   c[0] = v[0]->ui[4];
   c[1] = v[1]->ui[4];
   c[2] = v[2]->ui[4];
#endif


#if (IND & (SAVAGE_TWOSIDE_BIT | SAVAGE_OFFSET_BIT))
   {
      GLfloat ex = v[0]->v.x - v[2]->v.x;
      GLfloat ey = v[0]->v.y - v[2]->v.y;
      GLfloat fx = v[1]->v.x - v[2]->v.x;
      GLfloat fy = v[1]->v.y - v[2]->v.y;
      GLfloat cc  = ex*fy - ey*fx;

#if (IND & SAVAGE_TWOSIDE_BIT)
      {
	 GLuint  facing        = (cc > 0.0) ^ ctx->Polygon.FrontBit;
	 GLubyte (*vbcolor)[4] = VB->Color[facing]->data;
         SAVAGE_COLOR((char *)&v[0]->ui[4], vbcolor[e0]);
         SAVAGE_COLOR((char *)&v[1]->ui[4], vbcolor[e1]);
         SAVAGE_COLOR((char *)&v[2]->ui[4], vbcolor[e2]);
	 
      }
#endif

#if (IND & SAVAGE_OFFSET_BIT)
      {
	 offset = ctx->Polygon.OffsetUnits * 1.0/0x10000;
	 z[0] = v[0]->v.z;
	 z[1] = v[1]->v.z;
	 z[2] = v[2]->v.z;
	 if (cc * cc > 1e-16) {
	    GLfloat ez     = z[0] - z[2];
	    GLfloat fz     = z[1] - z[2];
	    GLfloat a      = ey*fz - ez*fy;
	    GLfloat b      = ez*fx - ex*fz;
	    GLfloat ic     = 1.0 / cc;
	    GLfloat ac     = a * ic;
	    GLfloat bc     = b * ic;
	    if (ac < 0.0f) ac = -ac;
	    if (bc < 0.0f) bc = -bc;
	    offset += MAX2(ac, bc) * ctx->Polygon.OffsetFactor;
	 }
	 v[0]->v.z += offset;
	 v[1]->v.z += offset;
	 v[2]->v.z += offset;
      }
#endif
   }
#endif

   savage_draw_triangle( savagectx, v[0], v[1], v[2] );

#if (IND & SAVAGE_OFFSET_BIT)
   v[0]->v.z = z[0];
   v[1]->v.z = z[1];
   v[2]->v.z = z[2];
#endif

#if ( IND & SAVAGE_TWOSIDE_BIT)
   v[0]->ui[4] = c[0];
   v[1]->ui[4] = c[1];
   v[2]->ui[4] = c[2];    
#endif

}


static void TAG(quad)( GLcontext *ctx, GLuint v0,
		       GLuint v1, GLuint v2, GLuint v3, 
		       GLuint pv )
{
   TAG(triangle)( ctx, v0, v1, v3, pv );
   TAG(triangle)( ctx, v1, v2, v3, pv );
}

static void TAG(line)( GLcontext *ctx, GLuint v0, GLuint v1, GLuint pv )
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
   savageVertexPtr savageVB = SAVAGE_DRIVER_DATA(ctx->VB)->verts;

#if (IND & (SAVAGE_TWOSIDE_BIT|SAVAGE_OFFSET_BIT)) 
   savageVertex tmp0 = savageVB[v0];
   savageVertex tmp1 = savageVB[v1];

   if (IND & SAVAGE_TWOSIDE_BIT) {
      GLubyte (*vbcolor)[4] = ctx->VB->ColorPtr->data;
      
      SAVAGE_COLOR((char *)&tmp0.v.color,vbcolor[v0]);
      SAVAGE_COLOR((char *)&tmp1.v.color,vbcolor[v1]);
      
   }

   /* Relies on precomputed LineZoffset from vbrender.c
    */
   if (IND & SAVAGE_OFFSET_BIT) {
      GLfloat offset = ctx->LineZoffset * (1.0 / 0x10000); 
      tmp0.v.z += offset;
      tmp1.v.z += offset;
   }

   savage_draw_line( imesa, &tmp0, &tmp1 );
#else
   savage_draw_line( imesa, &savageVB[v0], &savageVB[v1] );
#endif
}


static void TAG(points)( GLcontext *ctx, GLuint first, GLuint last )
{
   savageContextPtr imesa = SAVAGE_CONTEXT( ctx );
   struct vertex_buffer *VB = ctx->VB;
   savageVertexPtr savageVB = SAVAGE_DRIVER_DATA(VB)->verts;
   GLfloat sz = ctx->Point.Size * .5;
   int i;

   /* Culling is disabled automatically via. the
    * ctx->Driver.ReducedPrimitiveChange() callback.  
    */
   
   for(i=first;i<last;i++) {
      if(VB->ClipMask[i]==0) {
	 if (IND & SAVAGE_TWOSIDE_BIT) {	
	    savageVertex tmp0 = savageVB[i];
	    if (IND & SAVAGE_TWOSIDE_BIT) {
	       GLubyte (*vbcolor)[4] = VB->ColorPtr->data;
	       SAVAGE_COLOR((char *)&tmp0.v.color, vbcolor[i]);
	    }
	    if (IND & SAVAGE_OFFSET_BIT) {
	       GLfloat offset = ctx->PointZoffset * (1.0 / 0x10000);
	       tmp0.v.z += offset;
	    }
	    savage_draw_point( imesa, &tmp0, sz );
	 } else
	    savage_draw_point( imesa, &savageVB[i], sz );
      }
   }
}



static void TAG(init)( void )
{
   tri_tab[IND] = TAG(triangle);
   quad_tab[IND] = TAG(quad);
   line_tab[IND] = TAG(line);
   points_tab[IND] = TAG(points);
}


#undef IND
#undef TAG
