/*
 * Copyright 2003 VIA Technologies, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * VIA AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef _S3DATA_H_
#define _S3DATA_H_

/* Stream processor register definitions */
#define S_1_K1K2K3	0x8198	/* Secondary Stream 1*/
#define S_1_K4K5K6	0x819c	/* Secondary Stream 1*/
#define S_1_K7KB	0x81e4	/* Secondary Stream 1*/
#define S_2_K1K2K3	0x81f0	/* Secondary Stream 2*/
#define S_2_K4K5K6	0x81f4	/* Secondary Stream 2*/
#define S_2_K7KB	0x8200	/* Secondary Stream 2*/

#define BCI_S_1_K1K2K3	0xA6	/* Secondary Stream 1*/
#define BCI_S_1_K4K5K6	0xA7	/* Secondary Stream 1*/
#define BCI_S_1_K7KB	0xB9	/* Secondary Stream 1*/
#define BCI_S_2_K1K2K3	0xBC	/* Secondary Stream 2*/
#define BCI_S_2_K4K5K6	0xBD	/* Secondary Stream 2*/
#define BCI_S_2_K7KB	0xC0	/* Secondary Stream 2*/


#define P_CONTROL	0x8180	/* Primary Stream Control register*/

#define	  CKEY_LOW	0x8184	/* Color/Chroma Key Control*/
#define	  CKEY_LOW_2	0x8188	/* Color/Chroma Key Control 2*/

#define S_CONTROL		0x8190	/* Secondary Stream Control*/

#define	  CKEY_HI	0x8194	/* Chroma Key Upper Bound*/
#define	  CKEY_HI_2	0x818c	/* Chroma Key Upper Bound 2*/

#define S_HK1K2		0x8198	/* Secondary Stream Stretch/Filter const*/

#define	  BLEND_CONTROL		0x81a0	/* Blend Control*/
#define	  BLEND_CONTROL_PARA	0x8190	/* Blend Control*/

#define P_0	    0x81c0  /* Primary Stream Frame Buffer Address 0*/
#define P_1	    0x81c4  /* Primary Stream Frame Buffer Address 1*/
#define P_STRIDE    0x81c8  /* Primary Stream Stride*/
#define	  LPB_DB    0x81cc  /* Double buffer and LPB Support*/
#define S_0	    0x81d0  /* Secondary Stream Frame Buffer Addr 0*/
#define S_1	    0x81d4  /* Secondary Stream Frame Buffer Addr 1*/

#define S_2		0x8308	/* Secondary Stream Frame Buffer Address 2*/
#define S_2_0		0x81bc	/* Secondary Stream 2 Frame Buffer Addr 0*/

#define S_STRIDE	0x81d8	/* Secondary Stream Stride*/
#define S_STRIDE_2	0x81cc
#define OPAQUE_CONTROL	  0x81dc  /* Opaque Overlay Control*/
#define OPAQUE_CONTROL_2	0x8180	/* Opaque Overlay Control 2*/

#define S_VK1		0x81e0	/* K1 Vertical Scale Factor*/
#define S_VK2		0x81e4	/* K2 Vertical Scale Factor*/

#define S_VDDA		0x81e8	/* DDA Vertical Accumulator Init Value*/

#define	  FIFO_CONTROL	0x81ec	/* Stream FIFO and RAS Controls*/
#define P_XY		0x81f0	/* Primary Stream Window Start Coord*/
#define P_WH		0x81f4	/* Primary Stream Window Size*/

#define S_XY		0x81f8	/* Secondary Stream Window Start Coord*/
#define S_XY_2		0x820c	/* Secondary Stream 2 Window Start Coord*/

#define S_WH		0x81fc	/* Secondary Stream Window Size*/
#define S_WH_2		0x8210	/* Secondary Stream 2 Window Size*/

#define COLOR_ADJUST_REG 0x819c /* Color Adjustment Register*/
#define S_0_U			0x8208	/* Secondary Stream FrameBuf Addr 0 U (YUV420)*/
#define S_0_V			0x820c	/* Secondary Stream FrameBuf Addr 0 V (YUV420)*/
#define S_STRIDE_UV		0x8218	/* Secondary Stream Stride U/V (YUV420)*/


/* Begin GX3 Additional Streams processor register definitions*/
/* Note: Some of these registers have different names on Trio3D*/
#define P_FIFO_MONITORING_0 0x8200  /* Primary Stream FIFO Monitoring 0*/
#define S_FIFO_MONITORING_0 0x8204  /* Secondary Stream FIFO Monitoring 0*/
#define P_FIFO_MONITORING_1 0x8210  /* Primary Stream FIFO Monitoring 1*/
#define S_FIFO_MONITORING_1 0x8214  /* Secondary Stream FIFO Monitoring 1*/
#define P_FRAME_BUFFER_SIZE 0x8300  /* Primary Stream Frame Buffer Size*/
#define S_FRAME_BUFFER_SIZE 0x8304  /* Secondary Stream Frame Buffer Size*/
/* End GX3 register definitions*/

#define P_Format_Shift	    24
#define P_Format_Mask	    (7L << P_Format_Shift)
#define P_RGB8		(0L << P_Format_Shift)
#define P_RGB32_AXRGB		(1L << P_Format_Shift)
#define P_RGB15		(3L << P_Format_Shift)
#define P_RGB16		(5L << P_Format_Shift)
#define P_RGB24		(6L << P_Format_Shift)
#define P_RGB32		(7L << P_Format_Shift)
#define P_Filter_Shift	    28
#define P_Filter_Mask	    (7L << P_Filter_Shift)
#define P_1x		(0L << P_Filter_Shift)
#define P_2x		(1L << P_Filter_Shift)
#define P_2xBiLinear	    (2L << P_Filter_Shift)

#define BVCr_Shift	0
#define BVCr_Mask	(255L << BVCr_Shift)
#define GUCb_Shift	8
#define GUCb_Mask	(255L << GUCb_Shift)
#define RY_Shift	16
#define RY_Mask		(255L << RY_Shift)
#define Compare_Shift	    24
#define Compare_Mask	    (255L << Compare_Shift)
#define CompareBits7	    (0L	 << Compare_Shift)
#define CompareBits6t7	    (1L	 << Compare_Shift)
#define CompareBits5t7	    (2L	 << Compare_Shift)
#define CompareBits4t7	    (3L	 << Compare_Shift)
#define CompareBits3t7	    (4L	 << Compare_Shift)
#define CompareBits2t7	    (5L	 << Compare_Shift)
#define CompareBits1t7	    (6L	 << Compare_Shift)
#define CompareBits0t7	    (7L	 << Compare_Shift)
#define KeyFrom_Shift	    28
#define KeyFrom_Mask	    (1L << KeyFrom_Shift)
#define KeyFromStream	    (0L << KeyFrom_Shift)
#define KeyFromCompare	    (1L << KeyFrom_Shift)
#define ValueColorKey	    (0L << (KeyFrom_Shift+1))
#define IndexColorKey	    (1L << (KeyFrom_Shift+1))

/*For M3/M5/GX2 streams processor style*/
#define KMS_Shift	29
#define WindowKeying	    (0L << KMS_Shift )
#define AlphaKeying	(1L << KMS_Shift )
#define ColorKeying	(2L << KMS_Shift )
#define ChromaKeying	    (3L << KMS_Shift )
#define KeySelectP	(0L << KMS_Shift+2)
#define KeySelectS	(1L << KMS_Shift+2)


#define HDDA_Shift	0
#define HDDA_Mask	(((1L << 12)-1) << HDDA_Shift)
#define VHDDA_Mask  (((1L << 13)-1) << HDDA_Shift)
#define S_HDSCALE_Shift		16
#define S_HDSCALE_4		(2L << S_HDSCALE_Shift)
#define S_HDSCALE_8		(3L << S_HDSCALE_Shift)
#define S_HDSCALE_16		(4L << S_HDSCALE_Shift)
#define S_HDSCALE_32		(5L << S_HDSCALE_Shift)
#define S_HDSCALE_64		(6L << S_HDSCALE_Shift)
#define S_Format_Shift	    24
#define S_Format_Mask	    (7L << S_Format_Shift)
#define S_UYVY			(0L << S_Format_Shift)
#define S_YCrCb422	(1L << S_Format_Shift)
#define S_YUV422	(2L << S_Format_Shift)
#define S_RGB15		(3L << S_Format_Shift)
#define S_YUV211	(4L << S_Format_Shift)
#define S_YUV420	(4L << S_Format_Shift)
#define S_RGB16		(5L << S_Format_Shift)
#define S_RGB24		(6L << S_Format_Shift)
#define S_RGB32		(7L << S_Format_Shift)
#define S_YVYU		(0L << S_Format_Shift)
#define S_YUV411	(8L << S_Format_Shift)
#define S_Filter_Shift	    28
#define S_Filter_Mask	    (7L << S_Filter_Shift)
#define S_1x		(0L << S_Filter_Shift)
#define S_Upto2x	(1L << S_Filter_Shift)
#define S_2xTo4x	(2L << S_Filter_Shift)
#define S_Beyond4x	(3L << S_Filter_Shift)

#define HighVCr_Shift	    0
#define HighVCr_Mask	    (255L << HighVCr_Shift)
#define HighUCb_Shift	    8
#define HighUCb_Mask	    (255L << HighUCb_Shift)
#define HighY_Shift	16
#define HighY_Mask	(255L << HighY_Shift)

#define HK1_Shift	0
#define HK1_Mask	(((1L << 11) - 1) << HK1_Shift)
#define VHK1_Mask	(((1L << 12) - 1) << HK1_Shift)
#define HK2_Shift	16
#define HK2_Mask	(((1L << 11) - 1) << HK2_Shift)
#define VHK2_Mask	(((1L << 12) - 1) << HK2_Shift)
#define VK1_Shift	0
#define VK1_Mask	(((1L << 11) - 1) << VK1_Shift)
#define VVK1_Mask	(((1L << 12) - 1) << VK1_Shift)
#define VK2_Shift	0
#define VK2_Mask	(((1L << 11) - 1) << VK2_Shift)
#define VVK2_Mask	(((1L << 12) - 1) << VK2_Shift)
#define VDDA_Shift	0
#define VDDA_Mask	(((1L << 12) - 1) << VDDA_Shift)
#define VVDDA_Mask  (((1L << 13) - 1) << VDDA_Shift)

#define Ks_Shift	2
#define Ks_Mask		(7L << Ks_Shift)
#define Kp_Shift	10
#define Kp_Mask		(7L << Kp_Shift)
#define Compose_Shift	    24
#define Compose_Mask	    (7L << Compose_Shift)
#define SOnP		(0L << Compose_Shift)
#define POnS		(1L << Compose_Shift)
#define Dissolve	(2L << Compose_Shift)
#define Fade		(3L << Compose_Shift)
#define AlphaBlend	(4L << Compose_Shift)
#define KeyOnP		(5L << Compose_Shift)
#define KeyOnS		(6L << Compose_Shift)

#define FifoAlloc_Shift	    0
#define FifoAlloc_Mask	    (31L << StrFifoAlloc_Shift)
#define FifoAlloc24_0	    (0L	 << StrFifoAlloc_Shift)
#define FifoAlloc16_8	    (8L	 << StrFifoAlloc_Shift)
#define FifoAlloc12_12	    (12L << StrFifoAlloc_Shift)
#define FifoAlloc8_16	    (16L << StrFifoAlloc_Shift)
#define FifoAlloc0_24	    (24L << StrFifoAlloc_Shift)

#define S_FifoThresh_Shift  5
#define S_FifoThresh_Mask   (31L << S_FifoThresh_Shift)

#define P_FifoThresh_Shift  10
#define P_FifoThresh_Mask   (31L << P_FifoThresh_Shift)

#define RASLowTime_Shift    15
#define RASLowTime_Mask	    (1L << RASLowTime_Shift)
#define RASLowTimeFromCR68  (0L << RASLowTime_Shift)
#define RASLowTime2_5	    (1L << RASLowTime_Shift)

#define RASPreCharge_Shift  16
#define RASPreCharge_Mask   (1L << RASPreCharge_Shift)
#define RASPreChargeFromCR68	(0L << RASPreCharge_Shift)
#define RASPreCharge1_5	    (1L << RASPreCharge_Shift)

#define RASInactive_Shift   17
#define RASInactive_Mask    (1L << RASInactive_Shift)
#define RASInactiveLow	    (0L << RASInactive_Shift)
#define RASInactiveHigh	    (1L << RASInactive_Shift)

#define MemoryCycle_Shift   18
#define MemoryCycle_Mask    (1L << MemoryCycle_Shift)
#define MemoryCycle2	    (0L << MemoryCycle_Shift)
#define MemoryCycle1	    (1L << MemoryCycle_Shift)

#define H_Shift		0
#define H_Mask		(0x07ffL << H_Shift)
#define W_Shift		16
#define W_Mask		(0x07ffL << W_Shift)

#define Y_Shift		0
#define Y_Mask		(0x07ffL << Y_Shift)
#define X_Shift		16
#define X_Mask		(0x07ffL << X_Shift)

#define P_Select_Shift	    0
#define P_Select_Mask	    (1L << P_Select_Shift)
#define P_Select0	(0L << P_Select_Shift)
#define P_Select1	(1L << P_Select_Shift)
#define S_Select_Shift	    1
#define S_Select_Mask	    (3L << S_Select_Shift)
#define S_Select0	(0L << S_Select_Shift)
#define S_Select1	(1L << S_Select_Shift)
#define S_Select00Or11	    (2L << S_Select_Shift)
#define S_Select01Or10	    (3L << S_Select_Shift)
#define L_Select_Shift	    4
#define L_Select_Mask	    (1L << L_Select_Shift)
#define L_Select0	(0L << L_Select_Shift)
#define L_Select1	(1L << L_Select_Shift)
#define L_SelWait_Shift	    5
#define L_SelWait_Mask	    (1L << L_SelWait_Shift)
#define L_SelWaitNo	(0L << L_SelWait_Shift)
#define L_SelWaitYes	    (1L << L_SelWait_Shift)
#define L_SelAutoToggle_Shift	6
#define L_SelAutoToggle_Mask	(1L << L_SelAutoToggle_Shift)
#define L_SelAutoToggleNo   (0L << L_SelAutoToggle_Shift)
#define L_SelAutoToggleYes  (1L << L_SelAutoToggle_Shift)
#define L_FramesToDrop_Shift	8
#define L_FramesToDrop_Mask (3L << L_FramesToDrop_Shift)

#define OpqStart_Shift	    3
#define OpqStart_Mask	    (((1L << 10) - 1) << OpqStart_Shift)
#define OpqEnd_Shift	    19
#define OpqEnd_Mask	(((1L << 10) - 1) << OpqEnd_Shift)
#define OpqTopSel_Shift	    30
#define OpqTopSel_Mask	    (1L << OpqTopSel_Shift)
#define OpqTopSelS_	(0L << OpqTopSel_Shift)
#define OpqTopSelP_	(1L << OpqTopSel_Shift)
#define OpqCtrl_Shift	    31
#define OpqCtrl_Mask	    (1L << OpqCtrl_Shift)
#define OpqDisabled	(0L << OpqCtrl_Shift)
#define OpqEnabled	(1L << OpqCtrl_Shift)

/* The following defines are for VL and PCI system configuration*/
#define K2V_SRD_LPB_MASK    0x03
#define K2V_SRD_LPB	0x00
#define K2V_SRD_FC	K2V_SRD_LPB_MASK
#define K2V_SRD_COMP	    0x02

#define K2V_CR5C_SRC_MASK   0x03
#define K2V_CR5C_SRC_DIGITIZER	0x02
#define K2V_CR5C_SRC_MPEG   0x01

#define K2V_SR1C_MASK	    0x03
#define K2V_SR1C_VL	0x01
#define K2V_SR1C_PCI	    0x02

/* Useful defines*/

#define DDOVER_ALPHA \
    (DDOVER_ALPHASRC | \
     DDOVER_ALPHASRCCONSTOVERRIDE | \
     DDOVER_ALPHADEST | \
     DDOVER_ALPHADESTCONSTOVERRIDE)

#define DDOVER_KEY \
    (DDOVER_KEYSRC | \
     DDOVER_KEYSRCOVERRIDE |\
     DDOVER_KEYDEST | \
     DDOVER_KEYDESTOVERRIDE)

#define KEYING		0x1
#define BLENDING	0x2
#define OPAQUEING	0x4

/* Useful macros*/

/*macros for Virge/VX*/
#define VHDDA(w0,w1)  (((2*(w0-1)-(w1-1)) << HDDA_Shift) & VHDDA_Mask )
#define VVDDA(h1)     (((1-h1)	      << VDDA_Shift) & VVDDA_Mask )
#define VHK1(w0)      (((w0 - 1)      << HK1_Shift)  & VHK1_Mask  )
#define VHK2(w0,w1)   (((w0 - w1)     << HK2_Shift)  & VHK2_Mask  )
#define VHK1K2(w0,w1) (VHK1(w0) | VHK2(w0, w1))
#define VVK1(h0)      (((h0 - 1)      << VK1_Shift)  & VVK1_Mask  )
#define VVK2(h0,h1)   (((h0 - h1)     << VK2_Shift)  & VVK2_Mask  )

#define HDDA(w0,w1)  (((2*(w0-1)-(w1-1)) << HDDA_Shift) & HDDA_Mask )
#define VDDA(h1)     (((1-h1)	     << VDDA_Shift) & VDDA_Mask )
#define HK1(w0)	     (((w0 - 1)	     << HK1_Shift)  & HK1_Mask	)
#define HK2(w0,w1)   (((w0 - w1)     << HK2_Shift)  & HK2_Mask	)
#define HK1K2(w0,w1) (HK1(w0) | HK2(w0, w1))
#define VK1(h0)	     (((h0 - 1)	     << VK1_Shift)  & VK1_Mask	)
#define VK2(h0,h1)   (((h0 - h1)     << VK2_Shift)  & VK2_Mask	)
#define XY(x,y)	     ((((x+1)<<X_Shift)&X_Mask) | (((y+1)<<Y_Shift)&Y_Mask))
#define WH(w,h)	     ((((w?(w-1):0)<<W_Shift)&W_Mask) | (((h)<<H_Shift)&H_Mask))

#define K1K2K3(K1, K2, K3)  ((K1 & 0x1ff) | ((K2 << 9) & 0x3fe00) | ((K3 << 18) & 0x7fc0000))
#define K4K5K6(K4, K5, K6)  K1K2K3(K4, K5, K6)
#define K7KB(K7, KB)	    ((K7 & 0x1ff) | ((KB << 9) & 0x1fffe00))
#define INTERPOLATION(src, dst) ((65536 * (src) / (dst)) | ((src) << 20))
#define INTERPOLATION2(src, dst, srcClip) ((65536 * (src) / (dst)) | ((srcClip) << 20))

#define IsRGB15(this) \
    ((this->dwRBitMask == 0x7c00) && \
     (this->dwGBitMask == 0x03e0) && \
     (this->dwBBitMask == 0x001f))

#define IsRGB16(this) \
    ((this->dwRBitMask == 0xf800) && \
     (this->dwGBitMask == 0x07e0) && \
     (this->dwBBitMask == 0x001f))

#define IsRGB24(this) \
    ((this->dwRBitMask == 0x00ff0000) && \
     (this->dwGBitMask == 0x0000ff00) && \
     (this->dwBBitMask == 0x000000ff))

#define IsRGB32(this) \
    ((this->dwRBitMask == 0x00ff0000) && \
     (this->dwGBitMask == 0x0000ff00) && \
     (this->dwBBitMask == 0x000000ff))

#define RGB15to32( c ) \
    (((c & 0x7c00L) << 9) | \
     ((c & 0x03e0L) << 6) | \
     ((c & 0x001fL) << 3))

#define RGB16to32( c ) \
    (((c & 0xf800L) << 8) | \
     ((c & 0x07e0L) << 5) | \
     ((c & 0x001fL) << 3))

#define GetValidDDPF(this, lpDPF) \
{if(this->dwFlags & DDRAWISURF_HASPIXELFORMAT ) \
     lpDPF = &this->lpGbl->ddpfSurface; \
 else \
     lpDPF = &sData->ddpf; \
}

/* Equates for Pixel Formatter (Video Engine) 868/968*/

#define INPUT_RGB8  0x00000000
#define INPUT_RGB15 0x00600000
#define INPUT_RGB16 0x00700000
#define INPUT_RGB32 0x00300000
#define INPUT_YUV422	0x00480000
#define INPUT_YCrCb422	0x00400000
#define INPUT_RAW   0x00500000
#define OUTPUT_RGB8 0x00000000
#define OUTPUT_RGB15	0x00060000
#define OUTPUT_RGB16	0x00070000
#define OUTPUT_RGB32	0x00030000
#define OUTPUT_YUY2 0x000C0000
#define OUTPUT_RAW  0x00050000

#define CSCENABLE   0x40000000
#define STRETCH	    0x00000000
#define SHRINK	    0x80000000
#define SCREEN	    0x00000000
#define HOST	    0x40000000
#define FILTERENABLE	0x80000000
#define BILINEAR    0x00000000
#define LINEAR02420 0x00004000
#define LINEAR12221 0x00008000

#define PF_BUSY	    0x80000000

#define PF_NOP	    0x00018080
#define PF_CONTROL  0x00018088
#define PF_DDA	    0x0001808C
#define PF_STEP	    0x00018090
#define PF_CROP	    0x00018094
#define PF_SRCADDR  0x00018098
#define PF_DSTADDR  0x0001809C

#define GE_CMD	    0x00008118

#define GE_BUSY	    0x00000200
#define GE_FIFO_EMPTY	0x00000400



/*---------------------------------------------------------------------------*/
/* B C I   E Q U A T E S*/
/*---------------------------------------------------------------------------*/

/*PCI Configuration Registers*/
#define PCI_VENDOR_ID		0x8000
/*#define PCI_DEVICE_ID		  0x8002*/
#define PCI_COMMAND	    0x8004
#define PCI_STATUS	    0x8006
#define PCI_CLASS_CODE		0x8008
#define PCI_LATENCY_TIMER	0x800D
#define PCI_BASS_ADDRESS_0	0x8010
#define PCI_SUBSYSTEM_ID	0x802C
#define PCI_BIOS_BASE_ADDRESS	    0x8030


/*GX3 Configuration and Status Registers*/
#define STATUS_WORD_0		0x48C60
#define STATUS_WORD_1		0x48C64
#ifndef STATUS_WORD_2
#define STATUS_WORD_2		0x48C08
#endif

#define COMMAND_QUEUE_THRESHOLD	    0x48C10
#define COMMAND_BUFFER_ADDRESS_SIZE 0x48C14
#define SHADOW_STATUS_ADDRESS	    0x48C0C

#define VERTEX_BUFFER_ADDR	0x48C20
#define BCI_POWER_MANAGEMENT	    0x48C24

#define TILEON		    0x00001

#define TILED_SURFACE_REGISTER_0    0x48C40
#define TILED_SURFACE_REGISTER_1    0x48C44
#define TILED_SURFACE_REGISTER_2    0x48C48
#define TILED_SURFACE_REGISTER_3    0x48C4C
#define TILED_SURFACE_REGISTER_4    0x48C50


#define TILED_SURFACE_REGISTER_5	0x48C50

#define PRIMARY_APERTURE_REGISTER_0	TILED_SURFACE_REGISTER_0
#define PRIMARY_APERTURE_REGISTER_1	TILED_SURFACE_REGISTER_1
#define PRIMARY_APERTURE_REGISTER_2	TILED_SURFACE_REGISTER_2
#define ZBUFFER_APERTURE_REGISTER	TILED_SURFACE_REGISTER_3
#define TEXTURE_APERTURE_REGISTER	TILED_SURFACE_REGISTER_4

#ifdef PT_REVA

/*the following are offsets from BaseAddr2*/
    #define PRIMARY_APERTURE_OFFSET_0	    0x0000000
    #define PRIMARY_APERTURE_OFFSET_1	    0x1000000
    #define PRIMARY_APERTURE_OFFSET_2	    0x2000000
    #define ZBUFFER_APERTURE_OFFSET	    0x3000000

/*the following are offsets from BaseAddr3  */
    #define TEXTURE_APERTURE_OFFSET	    0x0000000

    #define TEXTURE_APERTURE_OFFSET_4	     0x0000000

#else

#define PRIMARY_APERTURE_OFFSET_0	0x2000000
#define PRIMARY_APERTURE_OFFSET_1	0x3000000
#define PRIMARY_APERTURE_OFFSET_2	0x4000000
#define ZBUFFER_APERTURE_OFFSET		0x5000000
#define TEXTURE_APERTURE_OFFSET		0x6000000

#define TEXTURE_APERTURE_OFFSET_4   0x6000000

#endif

/*Status Word Register 0*/
/* Beware the idle bits for rev B are valid for 48c00 and shadow*/
/* For 48c60, the idle bits need to shift down 4 bits.*/
#define BCI_IDLE		0x02000000
#define ENGINE_3D_IDLE		0x04000000
#define ENGINE_2D_IDLE		0x08000000
#define MOTION_COMPENSATION_UNIT_IDLE	0x10000000
/*#define VIDEO_ENGINE_IDLE	  0x20000000*/
#define MEU_IDLE		0x40000000
#define PAGE_FLIP_PENDING	0x80000000


/*Status Word Register 1*/
#define EVENT_TAG_MASK		0x0000FFFF
#define TEXTURE_SURFACE_TAG_MASK    0xFFFF0000


/*Status Word Register 2*/
#define CURRENT_SCAN_LINE_MASK	    0x000007FF
#define VERTICAL_RETRACE_COUNT_MASK 0x0007F8FF
#define MPEG2_FRAME_NUMBER_MASK	    0x07F80000



/*Status Word Register 3*/
#define INTERRUPT_ID_MASK	0x0000FFFF


/*BCI Commands*/
#define DRAW_PRIMITIVE		0x80000000
#define DRAW_INDEXED_PRIMITIVE	    0x88000000
#define SET_REGISTER		0x90000000
#define UPDATE_SHADOW_STATUS	    0x98000000
#define QUEUED_PAGE_FLIP	0xA0000000
#define MASTERED_IMAGE_TRANSFER	    0xA8000000
#define SCALED_IMAGE_TRANSFER	    0xB0000000
#define MACROBLOCK	    0xB8000000
#define WAIT		    0xC0000000
#define COMMANDS_2D	    0x40000000
#define COMMANDS_3D	    0x80000000

/*Update Shadow Flags*/
#define UPDATE_SHADOW_AGPTEXDELAY	0x1
#define UPDATE_SHADOW_PALETTEUPDATE	0x2

/*Wait command*/
#define WAIT_3D_ENGINE_IDLE	0x00010000
#define WAIT_2D_ENGINE_IDLE	0x00020000
#define WAIT_MAU_IDLE		0x00040000
#define WAIT_VEU_IDLE		0x00080000
#define WAIT_MEU_MIT_IDLE	0x00100000
#define WAIT_MEU_SIT_IDLE	0x00200000
#define WAIT_MEU_VX_IDLE	0x00400000
#define WAIT_MEU_MAUT_IDLE	0x00800000
#define WAIT_SCANLINE		0x02000000
#define WAIT_SCANLINE_IGA1	0x01000000
#define WAIT_SCANLINE_IGA2	0x02000000

/*Set register command*/
#define HIGH_ENABLE	    0x04000000
#define LOW_ENABLE	    0x02000000
#define ONE_REGISTER		0X00010000
#define TWO_REGISTER		0X00020000
#define FIVE_REGISTER		0X00050000
#define SIX_REGISTER		0X00060000
#define SEVEN_REGISTER		0X00070000
#define EIGHT_REGISTER		0X00080000
#define ONE_DW_REG	    0X06010000


/*The UpdateShadowStatus*/
#define WAIT_TILL_SPECIFIED_SCAN    0x04000000


/*The DrawPrimitive Command and DrawIndexedPrimitive*/
#define TYPE_TRIANGLE_LIST	0x00000000
#define TYPE_TRIANGLE_STRIP	0x02000000
#define TYPE_TRIANGLE_FAN	0x04000000
#define TYPE_QUAD_LIST		0x06000000


/*The QueuedPageFlip Command*/
#define WAIT_VSYNC	    0x04000000
#define MIU_SYNC	    0x00800000
#define PRI_SEC		    0x00400000



/*The Master Image Transfer Command*/
#define CTG_BLOCK	    0x04000000
#define NON_CTG_BLOCK		0x00000000
#define DIR_LOCAL_TO_SYSTEM	0x00400000
#define DIR_SYSTEM_TO_LOCAL	0x00000000

#define SYSTEM_MEM	    0x00000001
#define AGP_MEM		    0x00000002
#define TILED_ADDRESS		0x00000004

/*Mastered Image Control Register*/
#define FROM_SYS_TO_LOCAL	0x00000000
#define FROM_LOCAL_TO_SYS	0x80000000
#define SYS_MEM_NOT_CONTIGUOUS	    0x00000000
#define SYS_MEM_CONTIGUOUS	0x20000000
#define BPP_1		    0x00000000
#define BPP_2		    0x10000000
#define BPP_4		    0x20000000
#define BPP_8		    0x30000000

#define COLORFILL_BLT_CMD	0x000C4AD4   /* Rectangle fill command*/
#define TRANSPARENT_BLT_CMD	0x000C4B03   /**/
#define SRCCOPY_BLT_CMD		0x000C4803
#define XY_POSITIVE	0x03000000

/*Scaler control*/

/*Color Format for Src*/
#define SRC_YCbCr420		(0x0<<25)
#define SRC_YCbCr	    (0x1<<25)
#define SRC_RGB565	    (0x2<<25)
#define SRC_XRGB8888		(0x3<<25)

/*Color Format for Dest*/
#define DEST_YCbCr	    (0x1<<23)
#define DEST_RGB565	    (0x2<<23)
#define DEST_XRGB8888		(0x3<<23)


/*Tag Flags*/
#define IMAGE_TRANSFER		0X0001
#define VERTEX_BUFFER		0X0002
#define TEXTURE_SURFACE		0X0004


/*Frame Descriptor*/
#define FRAME_I_PIC	    (0x1<<27)
#define FRAME_P_PIC	    (0x2<<27)
#define FRAME_B_PIC	    (0x3<<27)
#define FRAME_TOP_FILED		(0x1<<29)
#define FRAME_BOTTOM_FILED	(0x2<<29)
#define FRAME_FRAME	    (0x3<<29)
#define FRAME_FIRST_P_FILED	0x0
#define FRAME_SECOND_P_FILED	    (0x1<<31)


/*Macroblock descriptor*/
#define MB_DECODE_CrBLK		0x1
#define MB_DECODE_CbBLK		(0x1<<1)
#define MB_DECODE_Y3BLK		(0x1<<2)
#define MB_DECODE_Y2BLK		(0x1<<3)
#define MB_DECODE_Y1BLK		(0x1<<4)
#define MB_DECODE_Y0BLK		(0x1<<5)
#define MB_DCT_FRAME		0x0
#define MB_DCT_FIELD		(0x1<<6)
#define MB_TYPE_INTRA		0x0
#define MB_TYPE_INTER_F		(0x1<<7)
#define MB_TYPE_INTER_B		(0x2<<7)
#define MB_TYPE_INTER_FB	(0x3<<7)
#define MB_MOTION_FIELD		(0x1<<9)
#define MB_MOTION_FRAME		(0x2<<9)
#define MB_MOTION_DP		(0x3<<9)
#define MB_MV3_DISABLED		0x0
#define MB_MV3_ENABLED		(0x1<<23)
#define MB_MV2_DISABLED		0x0
#define MB_MV2_ENABLED		(0x1<<24)
#define MB_MV1_DISABLED		0x0
#define MB_MV1_ENABLED		(0x1<<25)
#define MB_MV0_DISABLED		0x0
#define MB_MV0_ENABLED		(0x1<<26)
/* Motion Vector descriptor
#define MV_TOP_REF_FIELD	0x0
#ifndef MV_BOTTOM_REF_FIELD
#define MV_BOTTOM_REF_FIELD	(0x1<<28)
#endif
*/
/*Macroblock Data Address*/
#define FRAME_BUFFER		0x0
#define SYSTEM_MEMORY		0x1
#define SYSTEM_PCI	    0x0
#define SYSTEM_AGP	    0x2

/*Block Count Register*/
#define PARTIAL_FRAME_ENABLE	    0x1<31

#define MC_NEW_PICTURE		0x2
#define MC_SCALED	    0x4

/*2D Command for BCI interface*/
#define CMD_RECTANGLE		(0x1<<27)
#define CMD_SCANLINE		(0x2<<27)
#define CMD_B_LINE	    (0x3<<27)
#define CMD_BYTE_TEXT		(0x4<<27)
#define CMD_BIT_TEXT		(0x5<<27)
#define CMD_TEXT_W95		(0x0<<26)
#define CMD_TEXT_WNT		(0x1<<26)
#define CMD_DRAW_LAST_PIX	(0x0<<26)
#define CMD_NOT_DRAW_LAST_PIX	    (0x1<<26)

#define CMD_BLT_B_TO_T		(0x0<<25)
#define CMD_BLT_T_TO_B		(0x1<<25)
#define CMD_BLT_R_TO_L		(0x0<<24)
#define CMD_BLT_L_TO_R		(0x1<<24)

#define MIX_DEST_BLACK		(0x00<<16)
#define MIX_DEST_DN	    (0x55<<16)
#define MIX_DEST_D	    (0xAA<<16)
#define MIX_DEST_WHITE		(0xFF<<16)

#define MIX_SD_DSON	    (0x11<<16)
#define MIX_SD_DSNA	    (0x22<<16)
#define MIX_SD_SN	    (0x33<<16)
#define MIX_SD_SDNA	    (0x44<<16)
#define MIX_SD_DSX	    (0x66<<16)
#define MIX_SD_DSAN	    (0x77<<16)
#define MIX_SD_DSA	    (0x88<<16)
#define MIX_SD_DSXN	    (0x99<<16)
#define MIX_SD_DSNO	    (0xBB<<16)
#define MIX_SD_S	    (0xCC<<16)
#define MIX_SD_SDNO	    (0xDD<<16)
#define MIX_SD_DSO	    (0xEE<<16)

#define MIX_PD_DPON	    (0x05<<16)
#define MIX_PD_DPNA	    (0x0A<<16)
#define MIX_PD_DPN	    (0x0F<<16)
#define MIX_PD_PDNA	    (0x50<<16)
#define MIX_PD_PN	    (0x55<<16)
#define MIX_PD_DPX	    (0x5A<<16)
#define MIX_PD_DPAN	    (0x5F<<16)
#define MIX_PD_DPA	    (0xA0<<16)
#define MIX_PD_PDXN	    (0xA5<<16)
#define MIX_PD_PDNO	    (0xAF<<16)
#define MIX_PD_P	    (0xF0<<16)
#define MIX_PD_DDNO	    (0xF5<<16)
#define MIX_PD_DPO	    (0xFA<<16)

#define CLR_CURRENT	    (0x0<<15)
#define CLR_SPECIFIED		(0x1<<15)

#define DC_NO_CLIPPING		(0x0<<13)
#define DC_CLIP_CURRENT		(0x1<<13)
#define DC_CLIP_LEFT_RIGHT	(0x2<<13)
#define DC_CLIP_SPECIFIED	(0x3<<13)

#define DST_CURRENT_GBD		(0x0<<10)
#define DST_RESERVED		(0x1<<10)
#define DST_CURRENT_PBD		(0x2<<10)
#define DST_SPECIFIED_PBD	(0x3<<10)
#define DST_CURRENT_SBD		(0x4<<10)
#define DST_SPECIFIED_SBD	(0x5<<10)

#define ST_SCC_ENABLED		(0x1<<9)

#define SRC_SOLID_COLOR		(0x0<<5)
#define SRC_CURRENT_GBD		(0x1<<5)
#define SRC_HOST_COLOR		(0x2<<5)
#define SRC_HOST_MONO		(0x3<<5)
#define SRC_CURRENT_PBD_COLOR	    (0x4<<5)
#define SRC_CURRENT_PBD_MONO	    (0x5<<5)
#define SRC_SPECIFIED_PBD_COLOR	    (0x6<<5)
#define SRC_SPECIFIED_PBD_MONO	    (0x7<<5)
#define SRC_CURRENT_SBD_COLOR	    (0x8<<5)
#define SRC_CURRENT_SBD_MONO	    (0x9<<5)
#define SRC_SPECIFIED_SBD_COLOR	    (0xA<<5)
#define SRC_SPECIFIED_SBD_MONO	    (0xB<<5)

#define PT_SCC_ENABLED		(0x1<<4)

#define PAT_NO_PATTERN		0x0
#define PAT_RESERVED		0x1
#define PAT_HOST_COLOR		0x2
#define PAT_HOST_MONO		0x3
#define PAT_CURRENT_PBD_COLOR	    0x4
#define PAT_CURRENT_PBD_MONO	    0x5
#define PAT_SPECIFIED_PBD_COLOR	    0x6
#define PAT_SPECIFIED_PBD_MONO	    0x7
#define PAT_CURRENT_SBD_COLOR	    0x8
#define PAT_CURRENT_SBD_MONO	    0x9
#define PAT_SPECIFIED_SBD_COLOR	    0xA
#define PAT_SPECIFIED_SBD_MONO	    0xB

/*2D Command for MMIO and SetRegister interface*/
#define SR_CMD_PATBLT		0xE001	     /* 2D pat blt*/
#define SR_CMD_BITBLT		0xC001	     /* 2D bitBlt*/
#define SR_CMD_RECTFILL		0x4001	     /* 2D rectangle fill*/
#define SR_CMD_DRAWLINE		0x2001	     /* 2D line draw*/
#define SR_CMD_NOP	    0x0001
#define SR_LOWBYTE_FIRST	0x1000
#define SR_BUSSIZE_8		0x0000
#define SR_BUSSIZE_16		0x0200
#define SR_BUSSIZE_32		0x0400
#define SR_WAIT_YES	    0x0100
#define SR_LEFT_TO_RIGHT	0x0020
#define SR_TOP_TO_BOTTOM	0x0080
#define SR_DRAW_YES	    0x0010
#define SR_LASTPIXEL_OFF	0x0002

/*TempBuffShare definition in S3THKINFO <s3_ddtk.h , ddhal32.h>*/
#define AGP_MEM_ALLOC	    0x01000000


/*For Bitmap Descriptor*/
#define BW_DISABLED	    (0x1<<28)
#define TILE_FORMAT	    (0x1<<24)



/*Register number for Set Register Command*/

/*For 3D engine*/
#define BCI_DRAW_LOCAL_CONTROL	      0x1E
#define BCI_TEXTURE_PALETTE_ADDR      0x1F
#define BCI_TEXTURE0_CTRL	      0x20
#define BCI_TEXTURE1_CTRL	      0x21
#define BCI_TEXTURE0_ADDR	      0x22
#define BCI_TEXTURE1_ADDR	      0x23
#define BCI_TEXTURE0_BLEND_CTRL	      0x24
#define BCI_TEXTURE1_BLEND_CTRL	      0x25
#define BCI_TEXTURE_TRANSPARENT_COLOR 0x26
#define BCI_TEXTURE_DESC	      0x27
#define BCI_FOG_TBL_ENTRY0_3	      0x28
#define BCI_FOG_TBL_ENTRY4_7	      0x29
#define BCI_FOG_TBL_ENTRY8_11	      0x2A
#define BCI_FOG_TBL_ENTRY12_15	      0x2B
#define BCI_FOG_TBL_ENTRY16_19	      0x2C
#define BCI_FOG_TBL_ENTRY20_23	      0x2D
#define BCI_FOG_TBL_ENTRY24_27	      0x2E
#define BCI_FOG_TBL_ENTRY28_31	      0x2F
#define BCI_FOG_CONTROL		      0x30
#define BCI_STENCIL_CONTROL	      0x31
#define BCI_Z_CONTROL		      0x32
#define BCI_Z_OFFSET		      0x33
#define BCI_DEST_CONTROL	      0x34
#define BCI_DRAW_CONTROL0	      0x35
#define BCI_DRAW_CONTROL1	      0x36
#define BCI_Z_WATERMARKS	      0x37
#define BCI_3D_WATERMARKS	      0x38

/*For Motion Compensation Control*/
#define BCI_OLD_Y_BASE		0x40
#define BCI_OLD_Cb_BASE		0x41
#define BCI_OLD_Cr_BASE		0x42
#define BCI_NEW_Y_BASE		0x43
#define BCI_NEW_Cb_BASE		0x44
#define BCI_NEW_Cr_BASE		0x45
#define BCI_CURRENT_Y_BASE	0x46
#define BCI_CURRENT_Cb_BASE	0x47
#define BCI_CURRENT_Cr_BASE	0x48
#define BCI_FRAME_DESC		0X49

#define BCI_OLD_Y_Cb_BASE	   0x40
#define BCI_OLD_Cr_NEW_Y_BASE	   0x41
#define BCI_NEW_Cb_Cr_BASE	   0x42
#define BCI_CURRENT_Y_Cb_BASE	   0x43
#define BCI_CURRENT_Cr_FRAME_DESC  0X44
#define BCI_9B_IDCT		   0X45
#define BCI_MOTION_VECTOR_0	   0X46
#define BCI_MOTION_VECTOR_1	   0X47
#define BCI_MOTION_VECTOR_2	   0X48
#define BCI_MOTION_VECTOR_3	   0X49
#define BCI_MACROBLOCK_DESC	   0X4A
#define BCI_BLOCK_COUNT		   0X50
#define BCI_MACROBLOCK_DATA_ADDR   0X51
#define BCI_BLOCK_COUNT_PARA		0X50
#define BCI_MACROBLOCK_DATA_ADDR_PARA	0X4F

#define BCI_MIT			0x50
#define BCI_SOURCE_ADDR		0x51
#define BCI_DEST_ADDR		0x52


/*For Master Image Transfer*/
#define BCI_MASTERED_IMG_TX_SRC_ADDR	0x58
#define BCI_MASTERED_IMG_TX_DEST_ADDR	0x59
#define BCI_MASTERED_IMG_TX_SRC_XY  0x5A
#define BCI_MASTERED_IMG_TX_DEST_XY 0x5B
#define BCI_MASTERED_IMG_TX_DIMENSIONS	0x5C
#define BCI_MASTERED_IMG_TX_STRIDES 0x5D
#define BCI_MASTERED_IMG_TX_CONTROL 0x5E

/*For Pixel Formatter*/
#define BCI_PFM_CONTROL		    0x50
#define BCI_PFM_SRCADDR		    0x51
#define BCI_PFM_DESTADDR	    0x52
#define BCI_PFM_SRCDIMENSION	    0x53
#define BCI_PFM_DESTDIMENSION	    0x54
#define BCI_PFM_SRC_CBADDR	    0x55
#define BCI_PFM_SRC_CRADDR	    0x56
#define BCI_PFM_SRC_CRCBSTRIDE	    0x57


/*For Video engine unit*/
#define BCI_DVD_SUBPICTURE_ADDR	    0x60
#define BCI_DVD_SUBPICTURE_STRIDE   0x61
#define BCI_SRC_CRCB_STRIDE	0x62
#define BCI_SRC_DIMENSIONS	0x63
#define BCI_SRC_CR_ADDR		0x64
#define BCI_SRC_CB_ADDR		0x65
#define BCI_SRC_ADDR		0x66
#define BCI_DVD_SUBPICTURE_XY	    0x67
#define BCI_DVD_SUBPICTURE_WH	    0x68
#define BCI_VEU_COLOR_ADJUSTMENT    0x69
#define BCI_H_SCALE_RATIOS	0x6A
#define BCI_V_SCALE_RATIOS	0x6B
#define BCI_DEST_DIMENSIONS	0x6C
#define BCI_SRC_STRIDE		0x6D
#define BCI_SRC_XY	    0x6E
#define BCI_SCALER_CONTROL	0x6F
#define BCI_DVD_SUBPICTURE_HIGHLIGHT_XY 0x70
#define BCI_DVD_SUBPICTURE_HIGHLIGHT_WH 0x71

#define BCI_SUBPICTURE_COLOR_ENTRY_0	0x80
#define BCI_SUBPICTURE_COLOR_ENTRY_1	0x81
#define BCI_SUBPICTURE_COLOR_ENTRY_2	0x82
#define BCI_SUBPICTURE_COLOR_ENTRY_3	0x83
#define BCI_SUBPICTURE_COLOR_ENTRY_4	0x84
#define BCI_SUBPICTURE_COLOR_ENTRY_5	0x85
#define BCI_SUBPICTURE_COLOR_ENTRY_6	0x86
#define BCI_SUBPICTURE_COLOR_ENTRY_7	0x87
#define BCI_SUBPICTURE_COLOR_ENTRY_8	0x88
#define BCI_SUBPICTURE_COLOR_ENTRY_9	0x89
#define BCI_SUBPICTURE_COLOR_ENTRY_10	0x8A
#define BCI_SUBPICTURE_COLOR_ENTRY_11	0x8B
#define BCI_SUBPICTURE_COLOR_ENTRY_12	0x8C
#define BCI_SUBPICTURE_COLOR_ENTRY_13	0x8D
#define BCI_SUBPICTURE_COLOR_ENTRY_14	0x8E
#define BCI_SUBPICTURE_COLOR_ENTRY_15	0x8F

#define BCI_SUBPICTURE_HIGHLIGHT_COLOR_ENTRY_0	0x90
#define BCI_SUBPICTURE_HIGHLIGHT_COLOR_ENTRY_1	0x91
#define BCI_SUBPICTURE_HIGHLIGHT_COLOR_ENTRY_2	0x92
#define BCI_SUBPICTURE_HIGHLIGHT_COLOR_ENTRY_3	0x93

/*2D Packed Enhaced Registers*/
#define BCI_CURRENT_Y		0xD0
#define BCI_CURRENT_X		0xD0
#define BCI_DEST_Y	    0xD1
#define BCI_DEST_X	    0xD1
#define BCI_LINE_ERROR_TERM	0xD2
#define BCI_DRAWING_CMD		0xD3
#define BCI_SHORT_STROKE	0xD4
#define BCI_BACKGROUND_COLOR	    0xD5
#define BCI_FOREGROUND_COLOR	    0xD6
#define BCI_WRITE_MASK		0xD7
#define BCI_READ_MASK		0xD8
#define BCI_COLOR_COMPARE	0xD9
#define BCI_BACKGROUND_MIX	0xDA
#define BCI_FOREGROUND_MIX	0XDA
#define BCI_TOP_SCISSORS	0XDB
#define BCI_LEFT_SCISSORS	0XDB
#define BCI_BOTTOM_SCISSORS	0XDC
#define BCI_RIGHT_SCISSORS	0XDC
#define BCI_PIXEL_CONTROL	0XDD
#define BCI_MULTIFUNC_CTRL_MISC2    0XDD
#define BCI_MULTIFUNC_CTRL_MISC	    0XDE
#define BCI_READ_REG_SELECT	0XDE
#define BCI_MINOR_AXIS_PIXEL_COUNT  0XDF
#define BCI_MAJOR_AXIS_PIXEL_COUNT  0XDF
#define BCI_GLOBAL_BITMAP_DESC1	    0XE0
#define BCI_GLOBAL_BITMAP_DESC2	    0XE1
#define BCI_BURST_CMD_CTRL	0XE1
#define BCI_PRIMARY_BITMAP_DESC1    0XE2
#define BCI_PRIMARY_BITMAP_DESC2    0XE3
#define BCI_SECONDARY_BITMAP_DESC1  0XE4
#define BCI_SECONDARY_BITMAP_DESC2  0XE5

/*For streams processor*/
#define BCI_P_CONTROL		0xA0
#define BCI_CKEY_LOW		0xA1
#define BCI_S_CONTROL		0xA4
#define BCI_CKEY_HI	    0xA5
#define BCI_S_HK1K2	    0xA6
#define BCI_COLOR_CONTROL	0xA7

/*#define BCI_BLEND_CONTROL	  0xA8*/
#define BCI_CKEY_LOW_2		0xA2
#define BCI_CKEY_HI_2		0xA3
#define BCI_BLEND_CONTROL	0xA4
#define BCI_S_VDDA_1		0xBA
#define BCI_S_HDDA_1		0xA8
#define BCI_S_VDDA_2		0xC1
#define BCI_S_HDDA_2		0xA9
#define BCI_OPAQUE_CONTROL_2	0xA0
#define BCI_P_0			0xB0
#define BCI_S_XY_2	      0xC3
#define BCI_S_WH_2	      0xC4
#define BCI_S_STRIDE_2	      0xB3
#define BCI_P_1_BUFFER_SIZE	  0xC5
#define BCI_P_2_BUFFER_SIZE	0xC6
#define BCI_S_1_BUFFER_SIZE	  0xAA
#define BCI_S_2_BUFFER_SIZE	0xAB

#define BCI_P_1		    0xB1

/*Primary stream 2 address register indices*/
#define BCI_P_2_0	    0xAC
#define BCI_P_2_1	    0xAD
#define BCI_P_2_STRIDE	    0xAE

#define BCI_P_STRIDE		0xB2
#define BCI_LPB_DB	    0xB3
#define BCI_S_0		    0xB4
#define BCI_S_1		    0xB5
#define BCI_S_STRIDE		0xB6

/*Secondary stream 2 address register indices*/
#define BCI_S_2_0	    0xAF
#define BCI_S_2_1	    0xB8
#define BCI_S_2_2	    0xC2

#define BCI_OPAQUE_CONTROL	0xB7
#define BCI_S_VK1	    0xB8
#define BCI_S_VK2	    0xB9
#define BCI_S_SRC_LINE_COUNT	    0xBA
#define BCI_FIFO_CONTROL	0xBB
#define BCI_P_XY	    0xBC
#define BCI_P_WH	    0xBD
#define BCI_S_XY	    0xBE
#define BCI_S_WH	    0xBF
#define BCI_P_FIFO_MONITOR_0	    0xC0
#define BCI_S_FIFO_MONITOR_0	    0xC1


#define BCI_OPAQUE_CONTROL_2   0xA0

#define BCI_S_0_U			0xC2
#define BCI_S_0_V			0xC3
#define BCI_P_FIFO_MONITOR_1		0xC4
#define BCI_S_FIFO_MONOTOR_1		0xC5
#define BCI_S_STRIDE_UV			0xC6
#define BCI_P_BUFFER_SIZE_MX		0xC5
#define BCI_P_BUFFER_SIZE_S4		0xC7
#define BCI_S_BUFFER_SIZE		0xC8
#define BCI_S_2				0xC9

#define ALL_SURFACE	0x0
#define LEFT_HALF_SURFACE   0x1
#define RIGHT_HALF_SURFACE  0x2

/*Definition for Mediamatics decoder support of MC*/

/*PANEL expansion x factor*/
#define EXPAND_8_7	   0x30
#define EXPAND_5_4	   0x40
#define EXPAND_3_2	   0x60
#define EXPAND_2_1	   0x70
#define EXPAND_0000			0x00
#define EXPAND_0001			0x10
#define EXPAND_0010			0x20
#define EXPAND_0011			0x30
#define EXPAND_0100			0x40
#define EXPAND_0101	   0x50
#define EXPAND_0110			0x60
#define EXPAND_0111			0x70
#define EXPAND_1000			0x80
#define EXPAND_1001			0x90
#define EXPAND_1010			0xA0
#define EXPAND_1011			0xB0
#define EXPAND_1100			0xC0
#define EXPAND_1101			0xD0
#define EXPAND_1110			0xE0
#define EXPAND_1111			0xF0

/* Tile Format for Bitmap Descriptor*/
#define TILE_16_BPP	    0x2000000
#define TILE_32_BPP	    0x3000000
#endif /*_S3DATA_H_*/
